# -*- coding: UTF8 -*-
"""
some plotting functionality for different tasks
author: Michael Grupp

This file is part of evo (github.com/MichaelGrupp/evo).

evo is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

evo is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with evo.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import print_function  # Python 2.7 backwards compatibility

import os
import logging
import pickle
import collections
from enum import Enum

import matplotlib as mpl
from evo.tools.settings import SETTINGS

mpl.use(SETTINGS.plot_backend)
import matplotlib.cm as cm
import matplotlib.pyplot as plt
import mpl_toolkits.mplot3d.art3d as art3d
from mpl_toolkits.mplot3d import Axes3D
from matplotlib.collections import LineCollection
from matplotlib import rc

import numpy as np
import seaborn as sns

from evo.tools import user
from evo.core import trajectory

# configure matplotlib and seaborn according to package settings
sns.set(style=SETTINGS.plot_seaborn_style,
        palette=SETTINGS.plot_seaborn_palette,
        font=SETTINGS.plot_fontfamily,
        font_scale=SETTINGS.plot_fontscale
)


rc_params = {
    "lines.linewidth": SETTINGS.plot_linewidth,
    "text.usetex": SETTINGS.plot_usetex,
    "font.family": SETTINGS.plot_fontfamily,
    "font.serif": ['Cardo'],
    "pgf.texsystem": SETTINGS.plot_texsystem
}
mpl.rcParams.update(rc_params)

logger = logging.getLogger(__name__)


class PlotException(Exception):
    pass


class PlotMode(Enum):
    xy = "xy"
    xz = "xz"
    yx = "yx"
    yz = "yz"
    zx = "zx"
    zy = "zy"
    xyz = "xyz"


class PlotCollection:
    def __init__(self, title="", deserialize=None):
        self.title = " ".join(title.splitlines())  # one line title
        self.figures = collections.OrderedDict()  # remember placement order
        # hack to avoid premature garbage collection with Qt
        # (stackoverflow.com/questions/600289)
        self.root_window = None  # for now: init later in tabbed_qt_window
        if deserialize is not None:
            logger.debug("Deserializing PlotCollection from " + deserialize +
                         "...")
            self.figures = pickle.load(open(deserialize, 'rb'))

    def __str__(self):
        return self.title + " (" + str(len(self.figures)) + " figure(s))"

    def add_figure(self, name, fig):
        fig.tight_layout()
        self.figures[name] = fig

    def tabbed_qt4_window(self):
        from PyQt4 import QtGui
        from matplotlib.backends.backend_qt4agg import FigureCanvasQTAgg, NavigationToolbar2QT
        # mpl backend can already create instance
        # https://stackoverflow.com/a/40031190
        app = QtGui.QApplication.instance()
        if app is None:
            app = QtGui.QApplication([self.title])
        self.root_window = QtGui.QTabWidget()
        self.root_window.setWindowTitle(self.title)
        for name, fig in self.figures.items():
            tab = QtGui.QWidget(self.root_window)
            tab.canvas = FigureCanvasQTAgg(fig)
            vbox = QtGui.QVBoxLayout(tab)
            vbox.addWidget(tab.canvas)
            toolbar = NavigationToolbar2QT(tab.canvas, tab)
            vbox.addWidget(toolbar)
            tab.setLayout(vbox)
            for axes in fig.get_axes():
                if isinstance(axes, Axes3D):
                    # must explicitly allow mouse dragging for 3D plots
                    axes.mouse_init()
            self.root_window.addTab(tab, name)
        self.root_window.show()
        app.exec_()

    def tabbed_qt5_window(self):
        from PyQt5 import QtGui, QtWidgets
        from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg, NavigationToolbar2QT
        # mpl backend can already create instance
        # https://stackoverflow.com/a/40031190
        app = QtGui.QGuiApplication.instance()
        if app is None:
            app = QtWidgets.QApplication([self.title])
        self.root_window = QtWidgets.QTabWidget()
        self.root_window.setWindowTitle(self.title)
        for name, fig in self.figures.items():
            tab = QtWidgets.QWidget(self.root_window)
            tab.canvas = FigureCanvasQTAgg(fig)
            vbox = QtWidgets.QVBoxLayout(tab)
            vbox.addWidget(tab.canvas)
            toolbar = NavigationToolbar2QT(tab.canvas, tab)
            vbox.addWidget(toolbar)
            tab.setLayout(vbox)
            for axes in fig.get_axes():
                if isinstance(axes, Axes3D):
                    # must explicitly allow mouse dragging for 3D plots
                    axes.mouse_init()
            self.root_window.addTab(tab, name)
        self.root_window.show()
        app.exec_()

    def tabbed_tk_window(self):
        from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg, NavigationToolbar2Tk
        import sys
        if sys.version_info[0] < 3:
            import Tkinter as tkinter
            import ttk
        else:
            import tkinter
            from tkinter import ttk
        self.root_window = tkinter.Tk()
        self.root_window.title(self.title)
        # quit if the window is deleted
        self.root_window.protocol("WM_DELETE_WINDOW", self.root_window.quit)
        nb = ttk.Notebook(self.root_window)
        nb.grid(row=1, column=0, sticky='NESW')
        for name, fig in self.figures.items():
            fig.tight_layout()
            tab = ttk.Frame(nb)
            canvas = FigureCanvasTkAgg(self.figures[name], master=tab)
            canvas.draw()
            canvas.get_tk_widget().pack(side=tkinter.TOP, fill=tkinter.BOTH,
                                        expand=True)
            toolbar = NavigationToolbar2Tk(canvas, tab)
            toolbar.update()
            canvas._tkcanvas.pack(side=tkinter.TOP, fill=tkinter.BOTH,
                                  expand=True)
            for axes in fig.get_axes():
                if isinstance(axes, Axes3D):
                    # must explicitly allow mouse dragging for 3D plots
                    axes.mouse_init()
            nb.add(tab, text=name)
        nb.pack(side=tkinter.TOP, fill=tkinter.BOTH, expand=True)
        self.root_window.mainloop()
        self.root_window.destroy()

    def show(self):
        if len(self.figures.keys()) == 0:
            return
        if not SETTINGS.plot_split:
            if SETTINGS.plot_backend.lower() == "qt4agg":
                self.tabbed_qt4_window()
            elif SETTINGS.plot_backend.lower() == "qt5agg":
                self.tabbed_qt5_window()
            elif SETTINGS.plot_backend.lower() == "tkagg":
                self.tabbed_tk_window()
            else:
                plt.show()
        else:
            plt.show()

    def serialize(self, dest, confirm_overwrite=True):
        logger.debug("Serializing PlotCollection to " + dest + "...")
        if confirm_overwrite and not user.check_and_confirm_overwrite(dest):
            return
        else:
            pickle.dump(self.figures, open(dest, 'wb'))

    def export(self, file_path, confirm_overwrite=True):
        fmt = SETTINGS.plot_export_format.lower()
        if fmt == "pdf" and not SETTINGS.plot_split:
            if confirm_overwrite and not user.check_and_confirm_overwrite(
                    file_path):
                return
            import matplotlib.backends.backend_pdf
            pdf = matplotlib.backends.backend_pdf.PdfPages(file_path)
            for name, fig in self.figures.items():
                # fig.tight_layout()  # TODO
                pdf.savefig(fig)
            pdf.close()
            logger.info("Plots saved to " + file_path)
        else:
            for name, fig in self.figures.items():
                base, ext = os.path.splitext(file_path)
                dest = base + '_' + name + ext
                if confirm_overwrite and not user.check_and_confirm_overwrite(
                        dest):
                    return
                fig.tight_layout()
                fig.savefig(dest, fmt=fmt)
                logger.info("Plot saved to " + dest)


def set_aspect_equal_3d(ax):
    """
    kudos to https://stackoverflow.com/a/35126679
    :param ax: matplotlib 3D axes object
    """
    xlim = ax.get_xlim3d()
    ylim = ax.get_ylim3d()
    zlim = ax.get_zlim3d()

    from numpy import mean
    xmean = mean(xlim)
    ymean = mean(ylim)
    zmean = mean(zlim)

    plot_radius = max([
        abs(lim - mean_)
        for lims, mean_ in ((xlim, xmean), (ylim, ymean), (zlim, zmean))
        for lim in lims
    ])

    ax.set_xlim3d([xmean - plot_radius, xmean + plot_radius])
    ax.set_ylim3d([ymean - plot_radius, ymean + plot_radius])
    ax.set_zlim3d([zmean - plot_radius, zmean + plot_radius])


def prepare_axis(fig, plot_mode=PlotMode.xy, subplot_arg="111"):
    """
    prepares an axis according to the plot mode (for trajectory plotting)
    :param fig: matplotlib figure object
    :param plot_mode: PlotMode
    :param subplot_arg: optional if using subplots - the subplot id (e.g. '122')
    :return: the matplotlib axis
    """
    if plot_mode == PlotMode.xyz:
        ax = fig.add_subplot(subplot_arg, projection="3d", aspect="equal")
    else:
        ax = fig.add_subplot(subplot_arg, aspect="equal")
    plt.axis("equal")
    if plot_mode in {PlotMode.xy, PlotMode.xz, PlotMode.xyz}:
        xlabel = "$x$ (m)"
    elif plot_mode in {PlotMode.yz, PlotMode.yx}:
        xlabel = "$y$ (m)"
    else:
        xlabel = "$z$ (m)"
    if plot_mode in {PlotMode.xy, PlotMode.zy, PlotMode.xyz}:
        ylabel = "$y$ (m)"
    elif plot_mode in {PlotMode.zx, PlotMode.yx}:
        ylabel = "$x$ (m)"
    else:
        ylabel = "$z$ (m)"
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    if plot_mode == PlotMode.xyz:
        ax.set_zlabel('$z$ (m)')
    if SETTINGS.plot_invert_xaxis:
        plt.gca().invert_xaxis()
    if SETTINGS.plot_invert_yaxis:
        plt.gca().invert_yaxis()
    return ax


def plot_mode_to_idx(plot_mode):
    if plot_mode == PlotMode.xy or plot_mode == PlotMode.xyz:
        x_idx = 0
        y_idx = 1
    elif plot_mode == PlotMode.xz:
        x_idx = 0
        y_idx = 2
    elif plot_mode == PlotMode.yx:
        x_idx = 1
        y_idx = 0
    elif plot_mode == PlotMode.yz:
        x_idx = 1
        y_idx = 2
    elif plot_mode == PlotMode.zx:
        x_idx = 2
        y_idx = 0
    elif plot_mode == PlotMode.zy:
        x_idx = 2
        y_idx = 1
    z_idx = 2 if plot_mode == PlotMode.xyz else None
    return x_idx, y_idx, z_idx


def traj(ax, plot_mode, traj, style='-', color='black', label="", alpha=1.0):
    """
    plot a path/trajectory based on xyz coordinates into an axis
    :param ax: the matplotlib axis
    :param plot_mode: PlotMode
    :param traj: trajectory.PosePath3D or trajectory.PoseTrajectory3D object
    :param style: matplotlib line style
    :param color: matplotlib color
    :param label: label (for legend)
    :param alpha: alpha value for transparency
    """
    x_idx, y_idx, z_idx = plot_mode_to_idx(plot_mode)
    x = traj.positions_xyz[:, x_idx]
    y = traj.positions_xyz[:, y_idx]
    if plot_mode == PlotMode.xyz:
        z = traj.positions_xyz[:, z_idx]
        ax.plot(x, y, z, style, color=color, label=label, alpha=alpha)
        if SETTINGS.plot_xyz_realistic:
            set_aspect_equal_3d(ax)
    else:
        ax.plot(x, y, style, color=color, label=label, alpha=alpha)
    if label:
        ax.legend(frameon=True)


def colored_line_collection(xyz, colors, plot_mode=PlotMode.xy,
                            linestyles="solid"):
    if len(xyz) != len(colors):
        raise PlotException(
            "color values must have same length as xyz data: %d vs. %d" %
            (len(xyz), len(colors)))
    x_idx, y_idx, z_idx = plot_mode_to_idx(plot_mode)
    xs = [[x_1, x_2] for x_1, x_2 in zip(xyz[:-1, x_idx], xyz[1:, x_idx])]
    ys = [[x_1, x_2] for x_1, x_2 in zip(xyz[:-1, y_idx], xyz[1:, y_idx])]
    if plot_mode == PlotMode.xyz:
        zs = [[x_1, x_2] for x_1, x_2 in zip(xyz[:-1, z_idx], xyz[1:, z_idx])]
        segs = [list(zip(x, y, z)) for x, y, z in zip(xs, ys, zs)]
        line_collection = art3d.Line3DCollection(segs, colors=colors,
                                                 linestyles=linestyles)
    else:
        segs = [list(zip(x, y)) for x, y in zip(xs, ys)]
        line_collection = LineCollection(segs, colors=colors,
                                         linestyle=linestyles)
    return line_collection


def traj_colormap(ax, traj, array, plot_mode, min_map, max_map, title=""):
    """
    color map a path/trajectory in xyz coordinates according to
    an array of values
    :param ax: plot axis
    :param traj: trajectory.PosePath3D or trajectory.PoseTrajectory3D object
    :param array: Nx1 array of values used for color mapping
    :param plot_mode: PlotMode
    :param min_map: lower bound value for color mapping
    :param max_map: upper bound value for color mapping
    :param title: plot title
    """
    pos = traj.positions_xyz
    norm = mpl.colors.Normalize(vmin=min_map, vmax=max_map, clip=True)
    mapper = cm.ScalarMappable(
        norm=norm,
        cmap=SETTINGS.plot_trajectory_cmap)  # cm.*_r is reversed cmap
    mapper.set_array(array)
    colors = [mapper.to_rgba(a) for a in array]
    line_collection = colored_line_collection(pos, colors, plot_mode)
    ax.add_collection(line_collection)
    if plot_mode == PlotMode.xyz:
        ax.set_zlim(
            np.amin(traj.positions_xyz[:, 2]),
            np.amax(traj.positions_xyz[:, 2]))
        if SETTINGS.plot_xyz_realistic:
            set_aspect_equal_3d(ax)
    fig = plt.gcf()
    cbar = fig.colorbar(
        mapper, ticks=[min_map, (max_map - (max_map - min_map) / 2), max_map])
    cbar.ax.set_yticklabels([
        "{0:0.3f}".format(min_map),
        "{0:0.3f}".format(max_map - (max_map - min_map) / 2),
        "{0:0.3f}".format(max_map)
    ])
    if title:
        ax.legend(frameon=True)
        plt.title(title)


def traj_xyz(axarr, traj, style='-', color='black', label="", alpha=1.0,
             start_timestamp=None):
    """
    plot a path/trajectory based on xyz coordinates into an axis
    :param axarr: an axis array (for x, y & z)
                  e.g. from 'fig, axarr = plt.subplots(3)'
    :param traj: trajectory.PosePath3D or trajectory.PoseTrajectory3D object
    :param style: matplotlib line style
    :param color: matplotlib color
    :param label: label (for legend)
    :param alpha: alpha value for transparency
    :param start_timestamp: optional start time of the reference
                            (for x-axis alignment)
    """
    if len(axarr) != 3:
        raise PlotException("expected an axis array with 3 subplots - got " +
                            str(len(axarr)))
    if isinstance(traj, trajectory.PoseTrajectory3D):
        x = traj.timestamps - (traj.timestamps[0]
                               if start_timestamp is None else start_timestamp)
        xlabel = "$t$ (s)"
    else:
        x = range(0, len(traj.positions_xyz))
        xlabel = "index"
    ylabels = ["$x$ (m)", "$y$ (m)", "$z$ (m)"]
    for i in range(0, 3):
        axarr[i].plot(x, traj.positions_xyz[:, i], style, color=color,
                      label=label, alpha=alpha)
        axarr[i].set_ylabel(ylabels[i])
    axarr[2].set_xlabel(xlabel)
    if label:
        axarr[0].legend(frameon=True)


def traj_rpy(axarr, traj, style='-', color='black', label="", alpha=1.0,
             start_timestamp=None):
    """
    plot a path/trajectory's Euler RPY angles into an axis
    :param axarr: an axis array (for R, P & Y)
                  e.g. from 'fig, axarr = plt.subplots(3)'
    :param traj: trajectory.PosePath3D or trajectory.PoseTrajectory3D object
    :param style: matplotlib line style
    :param color: matplotlib color
    :param label: label (for legend)
    :param alpha: alpha value for transparency
    :param start_timestamp: optional start time of the reference
                            (for x-axis alignment)
    """
    if len(axarr) != 3:
        raise PlotException("expected an axis array with 3 subplots - got " +
                            str(len(axarr)))
    if isinstance(traj, trajectory.PoseTrajectory3D):
        x = traj.timestamps - (traj.timestamps[0]
                               if start_timestamp is None else start_timestamp)
        xlabel = "$t$ (s)"
    else:
        x = range(0, len(traj.orientations_euler))
        xlabel = "index"
    ylabels = ["$roll$ (deg)", "$pitch$ (deg)", "$yaw$ (deg)"]
    for i in range(0, 3):
        axarr[i].plot(x, np.rad2deg(traj.orientations_euler[:, i]), style,
                      color=color, label=label, alpha=alpha)
        axarr[i].set_ylabel(ylabels[i])
    axarr[2].set_xlabel(xlabel)
    if label:
        axarr[0].legend(frameon=True)


def trajectories(fig, trajectories, plot_mode=PlotMode.xy, title="",
                 subplot_arg="111"):
    """
    high-level function for plotting multiple trajectories
    :param fig: matplotlib figure
    :param trajectories: instances or iterables of PosePath3D or derived
    - if it's a dictionary, the keys (names) will be used as labels
    :param plot_mode: e.g. plot.PlotMode.xy
    :param title: optional plot title
    :param subplot_arg: optional matplotlib subplot ID if used as subplot
    """
    ax = prepare_axis(fig, plot_mode)
    cmap_colors = None
    if SETTINGS.plot_multi_cmap.lower() != "none":
        cmap = getattr(cm, SETTINGS.plot_multi_cmap)
        cmap_colors = iter(cmap(np.linspace(0, 1, len(trajectories))))

    # helper function
    def draw(t, name=""):
        if cmap_colors is None:
            color = next(ax._get_lines.prop_cycler)['color']
        else:
            color = next(cmap_colors)
        if SETTINGS.plot_usetex:
            name = name.replace("_", "\\_")
        traj(ax, plot_mode, t, '-', color, name)

    if isinstance(trajectories, trajectory.PosePath3D):
        draw(trajectories)
    elif isinstance(trajectories, dict):
        for name, t in trajectories.items():
            draw(t, name)
    else:
        for t in trajectories:
            draw(t)


def error_array(fig, err_array, x_array=None, statistics=None, threshold=None, cumulative=False,
                color='grey', name="error", title="", y_min=-1, y_max=-1,
                xlabel="index", ylabel=None, subplot_arg='111', linestyle="-", marker=None):
    """
    high-level function for plotting raw error values of a metric
    :param fig: matplotlib figure
    :param err_array: an nx1 array of values
    :param x_array: an nx1 array of x-axis values
    :param statistics: optional dictionary of {metrics.StatisticsType: value}
    :param threshold: optional value for horizontal threshold line
    :param cumulative: set to True for cumulative plot
    :param name: optional name of the value array
    :param title: optional plot title
    :param y_max: optional y_max limit, used to have same scale in different plots
    :param xlabel: optional x-axis label
    :param ylabel: optional y-axis label
    :param subplot_arg: optional matplotlib subplot ID if used as subplot
    :param linestyle: matplotlib linestyle
    :param marker: optional matplotlib marker style for points
    :return: the matplotlib figure with the plot
    """
    from matplotlib.ticker import FormatStrFormatter
    ax = fig.add_subplot(subplot_arg)
    if cumulative:
        if x_array:
            ax.plot(x_array, np.cumsum(err_array),
                    linestyle=linestyle, marker=marker, color=color, label=name)
        else:
            ax.plot(np.cumsum(err_array),
                    linestyle=linestyle, marker=marker, color=color, label=name)
    else:
        ymax = max(err_array)
        ymax_idx = np.argmax(err_array)
        xmax = ymax_idx
        if x_array:
            ax.plot(x_array, err_array, linestyle=linestyle, marker=marker, color=color, label=name)
            ax.set_xlim(min(x_array), max(x_array))
            xmax = x_array[ymax_idx]
        else:
            ax.plot(err_array, linestyle=linestyle, marker=marker, color=color, label=name)
            ax.set_xlim(0, len(err_array))
            xmax = ymax_idx
        #ax.plot(xmax, ymax, 'ro',markersize=10)
        #ax.text(xmax*(1-0.1), ymax*(1+(1/(1+ymax)*0.05)), 'Max: {0:.5f}'.format(ymax))
    # Set y axis limit if given:
    if y_max is not -1:
        ymin, _ = ax.get_ylim()
        ax.set_ylim(ymin, y_max)
    # Set y axis limit if given:
    if y_min is not -1:
        _, ymax = ax.get_ylim()
        ax.set_ylim(y_min, ymax)

    if statistics is not None:
        for stat_name, value in statistics.items():
            color = next(ax._get_lines.prop_cycler)['color']
            if stat_name in {"mean", "median", "rmse"}:
                ax.axhline(y=value, color=color, linewidth=2.0, label=stat_name+" ({0:.5f})".format(value))
            if stat_name == "std" and "mean" in statistics:
                mean, std = statistics["mean"], statistics["std"]
                ax.axhspan(mean - std,
                           mean + std,
                           color='cyan',
                           alpha=0.1,
                           edgecolor='green',
                           facecolor='none',
                           fill=False,
                           hatch='//',
                           linestyle='--',
                           linewidth=0.5,
                           zorder=0,
                           label=stat_name+" (+/- {0:.5f})".format(std))
    if threshold is not None:
        ax.axhline(y=threshold, color='red', linestyle='dashed', linewidth=2.0, label="threshold")
    # Remove vertical grid lines, as they are confused with spikes in the data.
    ax.xaxis.grid(which='both', visible=False)
    ax.yaxis.grid(which='major', ls='--')
    plt.ylabel(ylabel if ylabel else name)
    plt.xlabel(xlabel)
    if title != "":
        plt.title(title)
    plt.legend(frameon=True)
    return fig



#!/usr/bin/env python

from __future__ import print_function
import copy
import os
import yaml
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.ticker import FuncFormatter
from pylab import setp
from shutil import copyfile, move, rmtree, copytree, copy2

from evo.core import trajectory, sync, metrics

Y_MAX_APE_TRANS=[0.3, 0.25, 0.35, 0.5, 0.36, 0.170, 0.16, 0.4, 0.175, 0.24, 0.7]
Y_MAX_RPE_TRANS=[0.028, 0.025, 0.091, 0.21, 0.07, 0.03, 0.04, 0.15, 0.04, 0.06, 0.17]
Y_MAX_RPE_ROT=[0.4, 0.6, 0.35, 1.0, 0.3, 0.6, 1.5, 1.25, 0.6, 1.0, 2.6]

def find_step_of_base(x, base):
    return base * 10**np.floor(np.log10(np.abs(float(x))))

def _set_boxplot_colors(boxplot_object, color):
    setp(boxplot_object['boxes'][0], color=color)
    setp(boxplot_object['caps'][0], color=color)
    setp(boxplot_object['caps'][1], color=color)
    setp(boxplot_object['whiskers'][0], color=color)
    setp(boxplot_object['whiskers'][1], color=color)
    #setp(boxplot_object['fliers'], color=color)
    setp(boxplot_object['medians'][0], color=color)

def draw_boxplot(axis, stats, position, idx_experiment):
    """
        bxpstats : list of dicts
          A list of dictionaries containing stats for each boxplot.
          Required keys are:

          - ``med``: The median (scalar float).

          - ``q1``: The first quartile (25th percentile) (scalar
            float).

          - ``q3``: The third quartile (75th percentile) (scalar
            float).

          - ``whislo``: Lower bound of the lower whisker (scalar
            float).

          - ``whishi``: Upper bound of the upper whisker (scalar
            float).

          Optional keys are:

          - ``mean``: The mean (scalar float). Needed if
            ``showmeans=True``.

          - ``fliers``: Data beyond the whiskers (sequence of floats).
            Needed if ``showfliers=True``.

          - ``cilo`` & ``cihi``: Lower and upper confidence intervals
            about the median. Needed if ``shownotches=True``.

          - ``label``: Name of the dataset (string). If available,
            this will be used a tick label for the boxplot

        positions : array-like, default = [1, 2, ..., n]
          Sets the positions of the boxes. The ticks and limits
          are automatically set to match the positions.
    """
    colors = ['blue', 'black', 'green', 'red', 'mangenta', 'cyan', 'orange']
    bxpstats = []
    bxpstats_a = dict()
    bxpstats_a['med'] = stats['median']
    bxpstats_a['q1'] = stats['q1']
    bxpstats_a['q3'] = stats['q3']
    bxpstats_a['whislo'] = stats['min']
    bxpstats_a['whishi'] = stats['max']
    bxpstats.append(bxpstats_a)
    pb = axis.bxp(bxpstats,
                  positions=position,
                  widths=0.8, vert=True,
                  showcaps=True, showbox=True, showfliers=False, )
    _set_boxplot_colors(pb, colors[idx_experiment])

def draw_rpe_boxplots(output_dir, stats, n_segments):
    """ Draw boxplots from stats:
        which is a list that contains:
            - pipeline type (string) (like S, SP or SPR):
                - "relative_errors":
                    - segment distance (float) (like 10 or 20 etc):
                        - "max"
                        - "min"
                        - "mean"
                        - "median"
                        - "q1"
                        - "q3"
                        - "rmse"
        This function iterates over the pipeline types, and for each pipeline type, it plots
        the metrics achieved for each segment length. So the boxplot has in x-axis
        the number of segments considered, and in y-axis one boxplot per pipeline.
                        """

    colors = ['blue', 'black', 'green', 'red', 'mangenta', 'cyan', 'orange']
    if isinstance(stats, dict):
        n_experiment = len(stats)
        spacing = 1

        # Precompute position of boxplots in plot.
        pos = np.arange(0, n_segments * (n_experiment + spacing), (n_experiment + spacing))

        # Init axes
        # Use different plotting config.
        plt.style.use('default')
        import matplotlib as mpl
        from matplotlib import rc
        import seaborn as sns
        sns.reset_orig()
        mpl.rcParams.update(mpl.rcParamsDefault)
        rc('font',**{'family':'serif','serif':['Cardo'],'size':16})
        rc('text', usetex=True)

        fig = plt.figure(figsize=(6,6))
        ax_pos = fig.add_subplot(211, ylabel='RPE translation [m]')
        ax_yaw = fig.add_subplot(212, ylabel='RPE rotation [deg]', xlabel='Distance travelled [m]')
        dummy_plots_pos = []
        dummy_plots_yaw = []

        idx_experiment = 0
        x_labels = []
        final_max_e_pos=0.0
        final_max_e_yaw=0.0
        segment_lengths = []
        for pipeline_key, errors in sorted(stats.iteritems()):
            # The dummy plots are used to create the legends.
            dummy_plot_pos = ax_pos.plot([1,1], '-', color=colors[idx_experiment])
            dummy_plots_pos.append(dummy_plot_pos[0])
            dummy_plot_yaw = ax_yaw.plot([1,1], '-', color=colors[idx_experiment])
            dummy_plots_yaw.append(dummy_plot_yaw[0])
            x_labels.append(pipeline_key)
            if isinstance(errors, dict):
                assert("relative_errors" in errors)
                # Check that we have the expected number of segments
                assert(n_segments == len(errors['relative_errors']))
                idx_segment = 0
                for segment_length, stats in sorted(errors["relative_errors"].iteritems(), key = lambda item: int(item[0])):
                    segment_lengths.append(segment_length)
                    # Find max value overall, to set max in y-axis
                    max_e_pos = stats["rpe_trans"]["max"]
                    max_e_yaw = stats["rpe_rot"]["max"]
                    #max_e_pos = 10.2+0.02
                    #max_e_yaw = 30.0
                    if max_e_pos > final_max_e_pos:
                        final_max_e_pos = max_e_pos
                    if max_e_yaw > final_max_e_yaw:
                        final_max_e_yaw = max_e_yaw
                    # Draw boxplot
                    draw_boxplot(ax_pos, stats["rpe_trans"], [idx_experiment + pos[idx_segment]],
                                 idx_experiment)
                    draw_boxplot(ax_yaw, stats["rpe_rot"], [idx_experiment + pos[idx_segment]],
                                idx_experiment)
                    idx_segment = idx_segment + 1
            else:
                raise Exception("\033[91mValue in stats should be a dict: " + errors + "\033[99m")
            idx_experiment = idx_experiment + 1

        # Create legend.
        ax_pos.legend(dummy_plots_yaw, x_labels, bbox_to_anchor=(0., 1.02, 1., .102),
                      loc=3, ncol=3, mode='expand', borderaxespad=0.)

        def _ax_formatting(ax, dummy_plots, final_max_e):
            ax.yaxis.grid(ls='--', color='0.7')
            ax.yaxis.set_major_formatter(FuncFormatter(lambda y, pos: '%.2f'%y))
            # ax.xaxis.grid(which='major', visible=True, ls=' ')
            # ax.xaxis.grid(which='minor', visible=False)
            #ax.xaxis.set_major_formatter(plt.NullFormatter())
            ax.set_xticks(pos + 0.5*n_experiment - 0.5)
            ax.set_xticklabels(segment_lengths)
            ax.set_xlim(xmin=pos[0] - 1, xmax=pos[-1] + n_experiment + 0.2)
            ax.set_ylim(ymin=0, ymax=final_max_e)
            # Set yticks every multiple of 5 or 1. (so a tick 0.03 is now 0.00, 0.05,
            # and if 1 then a tick 0.034 is 0.03, 0.04)
            yticks = np.arange(0, final_max_e, find_step_of_base(final_max_e/5, 5))
            if len(yticks) < 4: # 4 is the minimum of yticks that we want.
                ax.set_yticks(np.arange(0, final_max_e, find_step_of_base(final_max_e/5, 1)))
            else:
                ax.set_yticks(yticks)
            for p in dummy_plots:
                p.set_visible(False)

        # give some extra space for the plot...
        final_max_e_pos += 0.05*final_max_e_pos
        final_max_e_yaw += 0.05*final_max_e_yaw
        _ax_formatting(ax_pos, dummy_plots_pos, final_max_e_pos)
        _ax_formatting(ax_yaw, dummy_plots_yaw, final_max_e_yaw)

        fig.savefig(os.path.join(output_dir, 'traj_relative_errors_boxplots.eps'),
                    bbox_inches="tight", format="eps", dpi=1200)
    else:
        raise Exception("\033[91mStats should be a dict: " + stats + "\033[99m")

    # Restore plotting config.
    from evo.tools.settings import SETTINGS
    plt.style.use('seaborn')
    # configure matplotlib and seaborn according to package settings
    sns.set(style=SETTINGS.plot_seaborn_style,
            palette=SETTINGS.plot_seaborn_palette,
            font=SETTINGS.plot_fontfamily,
            font_scale=SETTINGS.plot_fontscale
           )

    rc_params = {
        "lines.linewidth": SETTINGS.plot_linewidth,
        "text.usetex": SETTINGS.plot_usetex,
        "font.family": SETTINGS.plot_fontfamily,
        "font.serif": ['Cardo'],
        "pgf.texsystem": SETTINGS.plot_texsystem
    }
    mpl.rcParams.update(rc_params)

def draw_ape_boxplots(stats, output_dir):
    """ Draw boxplots from stats:
        which is a list that contains:
            - dataset name (string) (like V1_01_easy, MH_01_easy etc):
                - pipeline type (string) (like S, SP or SPR):
                    - "absolute_errors":
                        - "max"
                        - "min"
                        - "mean"
                        - "median"
                        - "q1"
                        - "q3"
                        - "rmse"
        This function iterates over the pipeline types, and for each pipeline type, it plots
        the metrics achieved, as a boxplot. So the boxplot has in x-axis the dataset name,
        and in y-axis one boxplot per pipeline."""
    colors = ['blue', 'black', 'green', 'red', 'mangenta', 'cyan', 'orange']
    if isinstance(stats, dict):
        n_param_values = len(stats)
        n_pipeline_types = len(stats.values()[0])
        spacing = 1

        # Precompute position of boxplots in plot.
        pos = np.arange(0, n_param_values * (n_pipeline_types + spacing),
                        (n_pipeline_types + spacing))

        # Use different plotting config.
        plt.style.use('default')
        import matplotlib as mpl
        from matplotlib import rc
        import seaborn as sns
        sns.reset_orig()
        mpl.rcParams.update(mpl.rcParamsDefault)
        rc('font',**{'family':'serif','serif':['Cardo'],'size':16})
        rc('text', usetex=True)

        # Init axis
        fig = plt.figure(figsize=(14, 6))
        ax_pos = fig.add_subplot(111, ylabel='APE translation error [m]', xlabel="Dataset")
        legend_labels = []
        legend_handles = []
        # Draw legend.
        color_id = 0
        for pipeline_type, pipeline_stats in sorted(stats.values()[0].iteritems()):
            # The dummy plots are used to create the legends.
            dummy_plot_pos = ax_pos.plot([1,1], '-', color=colors[color_id])
            legend_labels.append(pipeline_type)
            legend_handles.append(dummy_plot_pos[0])
            color_id = color_id + 1

        idx_param_value = 0
        final_max_e_pos=0.50
        xtick_labels=[]
        pipelines_failed = dict()
        for dataset_name, pipeline_types in sorted(stats.iteritems()):
            xtick_labels.append(dataset_name.replace('_', '\_'))
            if isinstance(pipeline_types, dict):
                idx_pipeline_type = 0
                for pipeline_type, pipeline_stats in sorted(pipeline_types.iteritems()):
                    if isinstance(pipeline_stats, dict):
                        # Find max value overall, to set max in y-axis
                        max_e_pos = pipeline_stats["absolute_errors"]["max"]
                        # if max_e_pos > final_max_e_pos:
                           # final_max_e_pos = max_e_pos
                        # Draw boxplot
                        draw_boxplot(ax_pos, pipeline_stats["absolute_errors"],
                                     [idx_pipeline_type + pos[idx_param_value]], idx_pipeline_type)
                    else:
                        # If pipeline_stats is not a dict, then it means the pipeline failed...
                        # Just plot a cross...
                        pipelines_failed[idx_pipeline_type] = [pipeline_type, idx_param_value]
                    idx_pipeline_type = idx_pipeline_type + 1
            else:
                raise Exception("\033[91mValue in stats should be a dict: " + errors + "\033[99m")
            idx_param_value = idx_param_value + 1

        # Draw crosses instead of boxplots for pipelines that failed.
        for idx_pipeline, pipeline_type_idx_param_pair in pipelines_failed.iteritems():
            x_middle = idx_pipeline + pos[pipeline_type_idx_param_pair[1]]
            x_1 = [x_middle - 0.5*spacing, x_middle + 0.5*spacing]
            y_1 = [0, final_max_e_pos]
            x_2 = [x_middle - 0.5*spacing, x_middle + 0.5*spacing]
            y_2 = [final_max_e_pos, 0]
            red_cross_plot = ax_pos.plot([1,1], 'xr')
            pipeline_type = pipeline_type_idx_param_pair[0]
            legend_labels.append("{} failure".format(pipeline_type))
            legend_handles.append(red_cross_plot[0])
            ax_pos.plot(x_1, y_1, '-r')
            ax_pos.plot(x_2, y_2, '-r')

        # Create legend.
        ax_pos.legend(legend_handles, legend_labels, bbox_to_anchor=(0., 1.02, 1., .102),
                      loc=3, ncol=3, mode='expand', borderaxespad=0.)

        def _ax_formatting(ax, dummy_plots, final_max_e):
            ax.yaxis.grid(ls='--', color='0.7')
            ax.yaxis.set_major_formatter(FuncFormatter(lambda y, pos: '%.2f'%y))
            # ax.xaxis.grid(which='major', visible=True, ls=' ')
            # ax.xaxis.grid(which='minor', visible=False)
            #ax.xaxis.set_major_formatter(plt.NullFormatter())
            ax.set_xticks(pos + 0.5*n_pipeline_types - 0.5)
            ax.set_xticklabels(xtick_labels, rotation=-40, ha='left')
            ax.set_xlim(xmin=pos[0] - 1, xmax=pos[-1] + n_pipeline_types + 0.2)
            ax.set_ylim(ymin=0, ymax= final_max_e)
            yticks = np.arange(0, final_max_e, find_step_of_base(final_max_e/5, 5))
            if len(yticks) < 4:
                ax.set_yticks(np.arange(0, final_max_e, find_step_of_base(final_max_e/5, 1)))
            else:
                ax.set_yticks(yticks)
            for p in dummy_plots:
                p.set_visible(False)

        # give some extra space for the plot...
        final_max_e_pos += 0.02
        _ax_formatting(ax_pos, legend_handles, final_max_e_pos)

        fig.savefig(os.path.join(output_dir, 'datasets_ape_boxplots.eps'),
                    bbox_inches="tight", format="eps", dpi=1200)
    else:
        raise Exception("\033[91mStats should be a dict: " + stats + "\033[99m")

    # Restore plotting config.
    from evo.tools.settings import SETTINGS
    plt.style.use('seaborn')
    # configure matplotlib and seaborn according to package settings
    sns.set(style=SETTINGS.plot_seaborn_style,
            palette=SETTINGS.plot_seaborn_palette,
            font=SETTINGS.plot_fontfamily,
            font_scale=SETTINGS.plot_fontscale
           )

    rc_params = {
        "lines.linewidth": SETTINGS.plot_linewidth,
        "text.usetex": SETTINGS.plot_usetex,
        "font.family": SETTINGS.plot_fontfamily,
        "font.serif": ['Cardo'],
        "pgf.texsystem": SETTINGS.plot_texsystem
    }
    mpl.rcParams.update(rc_params)

def draw_regression_simple_boxplot_APE(param_names, stats, output_dir, max_y = -1):
    """ Draw boxpots where x-axis are the values of the parameters in param_names, and the y-axis has boxplots with APE
    performance of the pipelines in stats.
    Stats is organized as follows:
        - param_value_dir (path to directory containing results for the parameter with given value)
            - pipeline (pipeline type e.g. S, SP or SPR)
                - results (which is actually -max, -min etc !OR! False if there are no results if the pipeline failed."""
    colors = ['blue', 'black', 'green', 'red', 'mangenta', 'cyan', 'orange']
    if isinstance(stats, dict):
        n_param_values = len(stats)
        assert(n_param_values > 0)
        n_pipeline_types = len(stats.values()[0])
        spacing = 1

        # Precompute position of boxplots in plot.
        pos = np.arange(0, n_param_values * (n_pipeline_types + spacing),
                        (n_pipeline_types + spacing))

        # Use different plotting config.
        plt.style.use('default')
        import matplotlib as mpl
        from matplotlib import rc
        import seaborn as sns
        sns.reset_orig()
        mpl.rcParams.update(mpl.rcParamsDefault)
        rc('font',**{'family':'serif','serif':['Cardo'],'size':16})
        rc('text', usetex=True)

        # Init axis
        fig = plt.figure(figsize=(6,2))
        param_names_dir = ""
        for i in param_names:
            param_names_dir += str(i) + "-"
            param_names_dir = param_names_dir[:-1]
        ax_pos = fig.add_subplot(111, ylabel='APE translation error [m]', xlabel="Values of parameter: {}".format(param_names_dir))
        legend_labels = []
        legend_handles = []
        # Draw legend.
        color_id = 0
        for pipeline_type, pipeline_stats in sorted(stats.values()[0].iteritems()):
            # The dummy plots are used to create the legends.
            dummy_plot_pos = ax_pos.plot([1,1], '-', color=colors[color_id])
            legend_labels.append(pipeline_type)
            legend_handles.append(dummy_plot_pos[0])
            color_id = color_id + 1

        idx_param_value = 0
        auto_scale = False
        final_max_e_pos = 0
        if max_y < 0:
            auto_scale = True
        else:
            final_max_e_pos = max_y
        param_values_boxplots=[]
        pipelines_failed = dict()
        for param_value_boxplots, pipeline_types in sorted(stats.iteritems()):
            param_values_boxplots.append(param_value_boxplots)
            if isinstance(pipeline_types, dict):
                idx_pipeline_type = 0
                for pipeline_type, pipeline_stats in sorted(pipeline_types.iteritems()):
                    if isinstance(pipeline_stats, dict):
                        # Find max value overall, to set max in y-axis
                        max_e_pos = pipeline_stats["absolute_errors"]["max"]
                        if auto_scale:
                            if max_e_pos > final_max_e_pos:
                               final_max_e_pos = max_e_pos
                        # Draw boxplot
                        draw_boxplot(ax_pos, pipeline_stats["absolute_errors"],
                                     [idx_pipeline_type + pos[idx_param_value]], idx_pipeline_type)
                    else:
                        # If pipeline_stats is not a dict, then it means the pipeline failed...
                        # Just plot a cross...
                        pipelines_failed[idx_pipeline_type] = [pipeline_type, idx_param_value]
                    idx_pipeline_type = idx_pipeline_type + 1
            else:
                raise Exception("\033[91mValue in stats should be a dict: " + errors + "\033[99m")
            idx_param_value = idx_param_value + 1

        # Draw crosses instead of boxplots for pipelines that failed.
        for idx_pipeline, pipeline_type_idx_param_pair in pipelines_failed.iteritems():
            x_middle = idx_pipeline + pos[pipeline_type_idx_param_pair[1]]
            x_1 = [x_middle - 0.5*spacing, x_middle + 0.5*spacing]
            y_1 = [0, final_max_e_pos]
            x_2 = [x_middle - 0.5*spacing, x_middle + 0.5*spacing]
            y_2 = [final_max_e_pos, 0]
            red_cross_plot = ax_pos.plot([1,1], 'xr')
            pipeline_type = pipeline_type_idx_param_pair[0]
            legend_labels.append("{} failure".format(pipeline_type))
            legend_handles.append(red_cross_plot[0])
            ax_pos.plot(x_1, y_1, '-r')
            ax_pos.plot(x_2, y_2, '-r')

        # Create legend.
        ax_pos.legend(legend_handles, legend_labels, bbox_to_anchor=(0., 1.02, 1., .102),
                      loc=3, ncol=3, mode='expand', borderaxespad=0.)

        def _ax_formatting(ax, dummy_plots, final_max_e):
            ax.yaxis.grid(ls='--', color='0.7')
            ax.yaxis.set_major_formatter(FuncFormatter(lambda y, pos: '%.2f'%y))
            # ax.xaxis.grid(which='major', visible=True, ls=' ')
            # ax.xaxis.grid(which='minor', visible=False)
            #ax.xaxis.set_major_formatter(plt.NullFormatter())
            ax.set_xticks(pos + 0.5*n_pipeline_types - 0.5)
            ax.set_xticklabels(param_values_boxplots)
            ax.set_xlim(xmin=pos[0] - 1, xmax=pos[-1] + n_pipeline_types + 0.2)
            ax.set_ylim(ymin=0, ymax= final_max_e)
            yticks = np.arange(0, final_max_e, find_step_of_base(final_max_e/5, 5))
            if len(yticks) < 4:
                ax.set_yticks(np.arange(0, final_max_e, find_step_of_base(final_max_e/5, 1)))
            else:
                ax.set_yticks(yticks)
            for p in dummy_plots:
                p.set_visible(False)

        # give some extra space for the plot...
        final_max_e_pos += 0.02
        _ax_formatting(ax_pos, legend_handles, final_max_e_pos)

        fig.savefig(os.path.join(output_dir, param_names_dir + '_absolute_errors_boxplots.eps'),
                    bbox_inches="tight", format="eps", dpi=1200)
    else:
        raise Exception("\033[91mStats should be a dict: " + stats + "\033[99m")

    # Restore plotting config.
    from evo.tools.settings import SETTINGS
    plt.style.use('seaborn')
    # configure matplotlib and seaborn according to package settings
    sns.set(style=SETTINGS.plot_seaborn_style,
            palette=SETTINGS.plot_seaborn_palette,
            font=SETTINGS.plot_fontfamily,
            font_scale=SETTINGS.plot_fontscale
           )

    rc_params = {
        "lines.linewidth": SETTINGS.plot_linewidth,
        "text.usetex": SETTINGS.plot_usetex,
        "font.family": SETTINGS.plot_fontfamily,
        "font.serif": ['Cardo'],
        "pgf.texsystem": SETTINGS.plot_texsystem
    }
    mpl.rcParams.update(rc_params)

def move_output_from_to(pipeline_output_dir, output_destination_dir):
    try:
        if (os.path.exists(output_destination_dir)):
            rmtree(output_destination_dir)
    except:
        print("Directory:" + output_destination_dir + " does not exist, we can safely move output.")
    try:
        if (os.path.isdir(pipeline_output_dir)):
            move(pipeline_output_dir, output_destination_dir)
        else:
            print("There is no output directory...")
    except:
        print("Could not move output from: " + pipeline_output_dir + " to: "
              + output_destination_dir)
        raise
    try:
        os.makedirs(pipeline_output_dir)
    except:
        print("Could not mkdir: " + pipeline_output_dir)
        raise

def ensure_dir(dir_path):
    """ Check if the path directory exists: if it does, returns true,
    if not creates the directory dir_path and returns if it was successful"""
    if not os.path.exists(dir_path):
        os.makedirs(dir_path)
    return True

def aggregate_ape_results(list_of_datasets, list_of_pipelines):
    RESULTS_DIR = '/home/tonirv/code/evo-1/results'
    # Load results.
    print("Loading dataset results")

    # Aggregate all stats for each pipeline and dataset
    stats = dict()
    for dataset_name in list_of_datasets:
        dataset_dir = os.path.join(RESULTS_DIR, dataset_name)
        stats[dataset_name] = dict()
        for pipeline_name in list_of_pipelines:
            pipeline_dir = os.path.join(dataset_dir, pipeline_name)
            # Get results.
            results_file = os.path.join(pipeline_dir, 'results.yaml')
            stats[dataset_name][pipeline_name]  = yaml.load(open(results_file,'r'))
            print("Check stats from " + results_file)
            checkStats(stats[dataset_name][pipeline_name])

    print("Drawing APE boxplots.")
    draw_ape_boxplots(stats, RESULTS_DIR)
    # Write APE table
    write_latex_table(stats, RESULTS_DIR)
    # Write APE table without S pipeline

def checkStats(stats):
    if not "relative_errors" in stats:
        print("Stats: ")
        print(stats)
        raise Exception("\033[91mWrong stats format: no relative_errors... \n"
                        "Are you sure you runned the pipeline and "
                        "saved the results? (--save_results).\033[99m")
    else:
        if len(stats["relative_errors"]) == 0:
            raise Exception("\033[91mNo relative errors available... \n"
                            "Are you sure you runned the pipeline and "
                            "saved the results? (--save_results).\033[99m")

        if not "rpe_rot" in stats["relative_errors"].values()[0]:
            print("Stats: ")
            print(stats)
            raise Exception("\033[91mWrong stats format: no rpe_rot... \n"
                            "Are you sure you runned the pipeline and "
                            "saved the results? (--save_results).\033[99m")
        if not "rpe_trans" in stats["relative_errors"].values()[0]:
            print("Stats: ")
            print(stats)
            raise Exception("\033[91mWrong stats format: no rpe_trans... \n"
                            "Are you sure you runned the pipeline and "
                            "saved the results? (--save_results).\033[99m")
    if not "absolute_errors" in stats:
        print("Stats: ")
        print(stats)
        raise Exception("\033[91mWrong stats format: no absolute_errors... \n"
                        "Are you sure you runned the pipeline and "
                        "saved the results? (--save_results).\033[99m")

def get_distance_from_start(gt_translation):
    distances = np.diff(gt_translation[:,0:3],axis=0)
    distances = np.sqrt(np.sum(np.multiply(distances,distances),1))
    distances = np.cumsum(distances)
    distances = np.concatenate(([0], distances))
    return distances

def locate_min(a):
    smallest = min(a)
    return smallest, [index for index, element in enumerate(a)
                      if smallest == element]

def write_latex_table(stats, results_dir):
    """ Write latex table with median, mean and rmse from stats:
            which is a list that contains:
                - dataset name (string) (like V1_01_easy, MH_01_easy etc):
                    - pipeline type (string) (like S, SP or SPR):
                        - "absolute_errors":
                            - "max"
                            - "min"
                            - "mean"
                            - "median"
                            - "q1"
                            - "q3"
                            - "rmse"
            This function iterates over the pipeline types, and for each pipeline type, it plots
            the metrics achieved, as a boxplot. So the boxplot has in x-axis the dataset name,
            and in y-axis one boxplot per pipeline."""
    # Tex stuff.
    # start_line = """\\begin{table}[H]
    # \\centering
    # \\resizebox{\\textwidth}{!}{
    # \\begin{tabular}{l p{1.4cm} p{1.4cm} p{1.4cm} p{1.4cm} p{1.4cm} p{1.4cm} p{1.4cm} p{1.4cm} p{1.4cm}}
    # \\hline
    # Sequence             & \\multicolumn{2}{c}{\\textbf{S}} & \\multicolumn{2}{c}{\\textbf{S + P}}  & \\multicolumn{2}{c}{\\textbf{S + P + R} (Proposed)}          \\\\ \\hline
                         # & Median APE Translation (m)  & Mean APE Translation (m) & RMSE APE Translation (m) &
                         # Median APE Translation (m)  & Mean APE Translation (m) & RMSE APE Translation (m) & Median
                         # APE Translation (m) & Mean APE Translation (m)  & RMSE APE translation (m) \\\\
    # """
    start_line = """\\begin{table}[H]
  \\centering
  \\caption{Accuracy of the state estimation when using Structureless and Projection factors (S + P), and our proposed approach using Structureless, Projection and Regularity factors (S + P + R)}
  \\label{tab:accuracy_comparison}
  \\begin{tabularx}{\\textwidth}{l *6{Y}}
    \\toprule
    & \\multicolumn{6}{c}{APE Translation} \\\\
    \\cmidrule{2-7}
    & \\multicolumn{3}{c}{\\textbf{S + P}}  & \\multicolumn{3}{c}{\\textbf{S + P + R} (Proposed)} \\\\
    \\cmidrule(r){2-4} \\cmidrule(l){5-7}
    Sequence & Median [cm] & Mean [cm] & RMSE [cm] & Median [cm] & Mean [cm] & RMSE [cm] \\\\
    \\midrule
    """

    end_line = """
    \\bottomrule
  \\end{tabularx}%
\\end{table}
"""
    bold_in = '& \\textbf{{'
    bold_out = '}} '
    end = '\\\\\n'

    all_lines = start_line

    winners = dict()
    for dataset_name, pipeline_types in sorted(stats.iteritems()):
        median_error_pos = []
        # mean_error_pos = []
        rmse_error_pos = []
        for pipeline_type, pipeline_stats in sorted(pipeline_types.iteritems()):
            # if pipeline_type is not "S": # Ignore S pipeline
            median_error_pos.append(pipeline_stats["absolute_errors"]["median"])
            # mean_error_pos.append(pipeline_stats["absolute_errors"]["mean"])
            rmse_error_pos.append(pipeline_stats["absolute_errors"]["rmse"])

        # Find winning pipeline
        _, median_idx_min = locate_min(median_error_pos)
        # _, mean_idx_min = locate_min(mean_error_pos)
        _, rmse_idx_min = locate_min(rmse_error_pos)

        # Store winning pipeline
        winners[dataset_name] = [median_idx_min,
                                 # mean_idx_min,
                                 rmse_idx_min]

    for dataset_name, pipeline_types in sorted(stats.iteritems()):
        start = '{:>25} '.format(dataset_name.replace('_', '\\_'))
        one_line = start
        pipeline_idx = 0
        for pipeline_type, pipeline_stats in sorted(pipeline_types.iteritems()):
            # if pipeline_type is not "S": # Ignore S pipeline
            median_error_pos = pipeline_stats["absolute_errors"]["median"] * 100 # as we report in cm
            # mean_error_pos = pipeline_stats["absolute_errors"]["mean"] * 100 # as we report in cm
            rmse_error_pos = pipeline_stats["absolute_errors"]["rmse"] * 100 # as we report in cm

            # Bold for min median error
            if len(winners[dataset_name][0]) == 1 and pipeline_idx == winners[dataset_name][0][0]:
                one_line += bold_in + '{:.1f}'.format(median_error_pos) + bold_out
            else:
                one_line += '& {:.1f} '.format(median_error_pos)

            # Bold for min mean error
            # if len(winners[dataset_name][1]) == 1 and winners[dataset_name][1][0] == pipeline_idx:
                # one_line += bold_in + '{:.1f}'.format(mean_error_pos) + bold_out
            # else:
                # one_line += '& {:.1f} '.format(mean_error_pos)

            # Bold for min rmse error
            # Do not bold, if multiple max
            if len(winners[dataset_name][1]) == 1 and winners[dataset_name][1][0] == pipeline_idx:
                one_line += bold_in + '{:.1f}'.format(rmse_error_pos) + bold_out
            else:
                one_line += '& {:.1f} '.format(rmse_error_pos)

            pipeline_idx += 1

        one_line += end
        all_lines += one_line
    all_lines += end_line

    # Save table
    results_file = os.path.join(results_dir, 'APE_table.tex')
    print("Saving table of APE results to: " + results_file)
    with open(results_file,'w') as outfile:
        outfile.write(all_lines)

def run_analysis(traj_ref_path, traj_est_path, segments, save_results, display_plot, save_plots,
                 save_folder, confirm_overwrite = False, idx_analysis = -1, discard_n_start_poses=0,
                discard_n_end_poses=0):
    """ Run analysis on given trajectories, saves plots on given path:
    :param traj_ref_path: path to the reference (ground truth) trajectory.
    :param traj_est_path: path to the estimated trajectory.
    :param save_results: saves APE, and RPE per segment results.
    :param save_plots: whether to save the plots.
    :param save_folder: where to save the plots.
    :param confirm_overwrite: whether to confirm overwriting plots or not.
    :param idx_analysis: optional param, to allow setting the same scale on different plots.
    """
    # Load trajectories.
    from evo.tools import file_interface
    traj_ref = file_interface.read_euroc_csv_trajectory(traj_ref_path)
    traj_est = file_interface.read_swe_csv_trajectory(traj_est_path)

    print("Registering trajectories")
    traj_ref, traj_est = sync.associate_trajectories(traj_ref, traj_est)

    print("Aligning trajectories")
    traj_est = trajectory.align_trajectory(traj_est, traj_ref, correct_scale = False,
                                           discard_n_start_poses = int(discard_n_start_poses),
                                           discard_n_end_poses = int(discard_n_end_poses))

    num_of_poses = traj_est.num_poses
    traj_est.reduce_to_ids(range(int(discard_n_start_poses), int(num_of_poses - discard_n_end_poses), 1))
    traj_ref.reduce_to_ids(range(int(discard_n_start_poses), int(num_of_poses - discard_n_end_poses), 1))

    results = dict()

    results["absolute_errors"] = dict()
    print("Calculating APE translation part")
    data = (traj_ref, traj_est)
    ape_metric = metrics.APE(metrics.PoseRelation.translation_part)
    ape_metric.process_data(data)
    ape_statistics = ape_metric.get_all_statistics()
    results["absolute_errors"] = ape_statistics
    print("mean:", ape_statistics["mean"])

    print("Calculating RPE translation part for plotting")
    rpe_metric_trans = metrics.RPE(metrics.PoseRelation.translation_part,
                                   1.0, metrics.Unit.frames, 0.0, False)
    rpe_metric_trans.process_data(data)
    rpe_stats_trans = rpe_metric_trans.get_all_statistics()
    print("mean:", rpe_stats_trans["mean"])

    print("Calculating RPE rotation angle for plotting")
    rpe_metric_rot = metrics.RPE(metrics.PoseRelation.rotation_angle_deg,
                                 1.0, metrics.Unit.frames, 1.0, False)
    rpe_metric_rot.process_data(data)
    rpe_stats_rot = rpe_metric_rot.get_all_statistics()
    print("mean:", rpe_stats_rot["mean"])

    results["relative_errors"] = dict()
    # Read segments file
    for segment in segments:
        results["relative_errors"][segment] = dict()
        print("RPE analysis of segment: " + segment)
        print("Calculating RPE segment translation part")
        rpe_segment_metric_trans = metrics.RPE(metrics.PoseRelation.translation_part,
                                       float(segment), metrics.Unit.meters, 0.01, True)
        rpe_segment_metric_trans.process_data(data)
        rpe_segment_stats_trans = rpe_segment_metric_trans.get_all_statistics()
        results["relative_errors"][segment]["rpe_trans"] = rpe_segment_stats_trans
        # print(rpe_segment_stats_trans)
        # print("mean:", rpe_segment_stats_trans["mean"])

        print("Calculating RPE segment rotation angle")
        rpe_segment_metric_rot = metrics.RPE(metrics.PoseRelation.rotation_angle_deg,
                                     float(segment), metrics.Unit.meters, 0.01, True)
        rpe_segment_metric_rot.process_data(data)
        rpe_segment_stats_rot = rpe_segment_metric_rot.get_all_statistics()
        results["relative_errors"][segment]["rpe_rot"] = rpe_segment_stats_rot
        # print(rpe_segment_stats_rot)
        # print("mean:", rpe_segment_stats_rot["mean"])

    if save_results:
        # Save results file
        results_file = os.path.join(save_folder, 'results.yaml')
        print("Saving analysis results to: " + results_file)
        if confirm_overwrite:
            if not user.check_and_confirm_overwrite(results_file):
                return
        with open(results_file,'w') as outfile:
            outfile.write(yaml.dump(results, default_flow_style=False))

    # For each segment in segments file
    # Calculate rpe with delta = segment in meters with all-pairs set to True
    # Calculate max, min, rmse, mean, median etc
    # Plot boxplot, or those cumulative figures you see in evo (like demographic plots)

    if display_plot or save_plots:
        print("loading plot modules")
        from evo.tools import plot

        print("plotting")
        plot_collection = plot.PlotCollection("Example")
        # metric values
        fig_1 = plt.figure(figsize=(8, 8))
        ymax = -1
        if idx_analysis is not -1:
            ymax = Y_MAX_APE_TRANS[idx_analysis]
        plot.error_array(fig_1, ape_metric.error, statistics=ape_statistics,
                         name="APE translation", title=""#str(ape_metric)
                         , xlabel="Keyframe index [-]",
                         ylabel="APE translation [m]", y_min= 0.0, y_max=ymax)
        plot_collection.add_figure("APE_translation", fig_1)

        # trajectory colormapped with error
        fig_2 = plt.figure(figsize=(8, 8))
        plot_mode = plot.PlotMode.xy
        ax = plot.prepare_axis(fig_2, plot_mode)
        plot.traj(ax, plot_mode, traj_ref, '--', 'gray', 'reference')
        plot.traj_colormap(ax, traj_est, ape_metric.error, plot_mode,
                           min_map=0.0, max_map=ymax,
                           title="APE translation error mapped onto trajectory [m]")
        plot_collection.add_figure("APE_translation_trajectory_error", fig_2)

        # RPE
        ## Trans
        ### metric values
        fig_3 = plt.figure(figsize=(8, 8))
        if idx_analysis is not -1:
            ymax = Y_MAX_RPE_TRANS[idx_analysis]
        plot.error_array(fig_3, rpe_metric_trans.error, statistics=rpe_stats_trans,
                         name="RPE translation", title=""#str(rpe_metric_trans)
                         , xlabel="Keyframe index [-]", ylabel="RPE translation [m]", y_max=ymax)
        plot_collection.add_figure("RPE_translation", fig_3)

        ### trajectory colormapped with error
        fig_4 = plt.figure(figsize=(8, 8))
        plot_mode = plot.PlotMode.xy
        ax = plot.prepare_axis(fig_4, plot_mode)
        traj_ref_trans = copy.deepcopy(traj_ref)
        traj_ref_trans.reduce_to_ids(rpe_metric_trans.delta_ids)
        traj_est_trans = copy.deepcopy(traj_est)
        traj_est_trans.reduce_to_ids(rpe_metric_trans.delta_ids)
        plot.traj(ax, plot_mode, traj_ref_trans, '--', 'gray', 'Reference')
        plot.traj_colormap(ax, traj_est_trans, rpe_metric_trans.error, plot_mode,
                           min_map=0.0, max_map=ymax,
                           title="RPE translation error mapped onto trajectory [m]"
                          )
        plot_collection.add_figure("RPE_translation_trajectory_error", fig_4)

        ## Rot
        ### metric values
        fig_5 = plt.figure(figsize=(8, 8))
        if idx_analysis is not -1:
            ymax = Y_MAX_RPE_ROT[idx_analysis]
        plot.error_array(fig_5, rpe_metric_rot.error, statistics=rpe_stats_rot,
                         name="RPE rotation error", title=""#str(rpe_metric_rot)
                         , xlabel="Keyframe index [-]", ylabel="RPE rotation [deg]", y_max=ymax)
        plot_collection.add_figure("RPE_rotation", fig_5)

        ### trajectory colormapped with error
        fig_6 = plt.figure(figsize=(8, 8))
        plot_mode = plot.PlotMode.xy
        ax = plot.prepare_axis(fig_6, plot_mode)
        traj_ref_rot = copy.deepcopy(traj_ref)
        traj_ref_rot.reduce_to_ids(rpe_metric_rot.delta_ids)
        traj_est_rot = copy.deepcopy(traj_est)
        traj_est_rot.reduce_to_ids(rpe_metric_rot.delta_ids)
        plot.traj(ax, plot_mode, traj_ref_rot, '--', 'gray', 'Reference')
        plot.traj_colormap(ax, traj_est_rot, rpe_metric_rot.error, plot_mode,
                           min_map=0.0, max_map=ymax,
                           title="RPE rotation error mapped onto trajectory [deg]")
        plot_collection.add_figure("RPE_rotation_trajectory_error", fig_6)

        if display_plot:
            print("Displaying plots.")
            plot_collection.show()

        if save_plots:
            print("Saving plots to: " + save_folder)
            plot_collection.export(save_folder + "/plots", False)

    ## Plot results
    #if args.plot or args.save_plot or args.serialize_plot:
        #    common.plot(
        #        args, result,
        #        result.trajectories[ref_name],
        #        result.trajectories[est_name])

    ## Save results
    #if args.save_results:
        #    logger.debug(SEP)
        #    if not SETTINGS.save_traj_in_zip:
            #        del result.trajectories[ref_name]
            #        del result.trajectories[est_name]
            #    file_interface.save_res_file(
            #        args.save_results, result, confirm_overwrite=not args.no_warnings)

# Run pipeline as a subprocess.
def run_vio(build_dir, dataset_dir, dataset_name, results_dir, pipeline_output_dir, pipeline_type,
           extra_flagfile_path = ""):
    """ Runs pipeline depending on the pipeline_type"""
    import subprocess
    return subprocess.call("{}/stereoVIOEuroc \
                           --logtostderr=1 --colorlogtostderr=1 --log_prefix=0 \
                           --dataset_path={}/{} --output_path={}\
                           --vio_params_path={}/params/{}/{} \
                           --tracker_params_path={}/params/{}/{} \
                           --flagfile={}/params/{}/{} --flagfile={}/params/{}/{} \
                           --flagfile={}/params/{}/{} --flagfile={}/params/{}/{} \
                           --flagfile={}/params/{}/{} --flagfile={}/params/{}/{} \
                           --log_output=True".format(
                               build_dir, dataset_dir, dataset_name, pipeline_output_dir,
                               results_dir, pipeline_type, "vioParameters.yaml",
                               results_dir, pipeline_type, "trackerParameters.yaml",
                               results_dir, pipeline_type, "flags/stereoVIOEuroc.flags",
                               results_dir, pipeline_type, "flags/Mesher.flags",
                               results_dir, pipeline_type, "flags/VioBackEnd.flags",
                               results_dir, pipeline_type, "flags/RegularVioBackEnd.flags",
                               results_dir, pipeline_type, "flags/Visualizer3D.flags",
                               results_dir, pipeline_type, extra_flagfile_path), \
                           shell=True)

def process_vio(build_dir, dataset_dir, dataset_name, results_dir, pipeline_output_dir,
                pipeline_type, SEGMENTS, save_results, plot, save_plots, output_file, run_pipeline, analyse_vio, discard_n_start_poses, discard_n_end_poses):
    """ build_dir: directory where the pipeline executable resides,
    dataset_dir: directory of the dataset,
    dataset_name: specific dataset to run,
    results_dir: directory where the results of the run will reside:
        used as results_dir/dataset_name/S, results_dir/dataset_name/SP, results_dir/dataset_name/SPR
        where each directory have traj_est.csv (the estimated trajectory), and plots if requested.
        results_dir/dataset_name/ must contain traj_gt.csv (the ground truth trajectory for analysis to work),
    pipeline_output_dir: where to store all output_* files produced by the pipeline,
    pipeline_type: type of pipeline to process (1: S, 2: SP, 3: SPR)
    SEGMENTS: segments for RPE boxplots,
    save_results: saves APE, and RPE per segment results of the run,
    plot: whether to plot the APE/RPE results or not,
    save_plots: saves plots of APE/RPE,
    output_file: the name of the trajectory estimate output of the vio which will then be copied as traj_est.csv,
    run_pipeline: whether to run the VIO to generate a new traj_est.csv,
    analyse_vio: whether to analyse traj_est.csv or not"""
    dataset_result_dir = results_dir + "/" + dataset_name + "/"
    dataset_pipeline_result_dir = dataset_result_dir + "/" + pipeline_type + "/"
    traj_ref_path = dataset_result_dir + "/traj_gt.csv"
    traj_est_s = dataset_result_dir + "/" + pipeline_type + "/" + "traj_es.csv"
    if run_pipeline:
        if run_vio(build_dir, dataset_dir, dataset_name, results_dir,
                   pipeline_output_dir, pipeline_type) == 0:
            print("Successful pipeline run.")
            print("\033[1mCopying output file: " + output_file + "\n to results file:\n" + \
                  traj_est_s + "\033[0m")
            copyfile(output_file, traj_est_s)
            try:
                output_destination_dir = dataset_pipeline_result_dir + "/output/"
                print("\033[1mCopying output dir: " + pipeline_output_dir
                      + "\n to destination:\n" + output_destination_dir + "\033[0m")
                move_output_from_to(pipeline_output_dir, output_destination_dir)
            except:
                print("\033[1mFailed copying output dir: " + pipeline_output_dir
                      + "\n to destination:\n" + output_destination_dir + "\033[0m")
        else:
            print("Pipeline failed on dataset: " + dataset_name)
            return False

    if analyse_vio:
        print("\033[1mAnalysing dataset: " + dataset_result_dir + " for pipeline "
              + pipeline_type + ".\033[0m")
        run_analysis(traj_ref_path, traj_est_s, SEGMENTS,
                     save_results, plot, save_plots, dataset_pipeline_result_dir, False,
                     return_id_of_dataset(dataset_name),
                     discard_n_start_poses,
                     discard_n_end_poses)
    return True

def return_id_of_dataset(dataset_name):
    if dataset_name == "MH_01_easy":
        return 0
    if dataset_name == "MH_02_easy":
        return 1
    if dataset_name == "MH_03_medium":
        return 2
    if dataset_name == "mh_04_difficult":
        return 3
    if dataset_name == "MH_05_difficult":
        return 4
    if dataset_name == "V1_01_easy":
        return 5
    if dataset_name == "V1_02_medium":
        return 6
    if dataset_name == "V1_03_difficult":
        return 7
    if dataset_name == "V2_01_easy":
        return 8
    if dataset_name == "V2_02_medium":
        return 9
    if dataset_name == "v2_03_difficult":
        return 10

def run_dataset(results_dir, dataset_dir, dataset_name, build_dir,
                run_pipeline, analyse_vio,
                plot, save_results, save_plots, save_boxplots, pipelines_to_run_list,
                discard_n_start_poses = 0, discard_n_end_poses = 0):
    """ Evaluates pipeline using Structureless(S), Structureless(S) + Projection(P), \
            and Structureless(S) + Projection(P) + Regular(R) factors \
            and then compiles a list of results """
    import time

    with open(os.path.join(results_dir, dataset_name, 'segments.txt'), 'r') as myfile:
        SEGMENTS = myfile.read().replace('\n', '').split(',')

    ################### RUN PIPELINE ################################
    pipeline_output_dir = results_dir + "/tmp_output/output"
    output_file = pipeline_output_dir + "/output_posesVIO.csv"
    has_a_pipeline_failed = False
    if len(pipelines_to_run_list) == 0:
        print("Not running pipeline...")
    for pipeline_type in pipelines_to_run_list:
        if process_vio(build_dir, dataset_dir, dataset_name, results_dir, pipeline_output_dir,
                       pipeline_type, SEGMENTS, save_results, plot, save_plots,
                       output_file, run_pipeline, analyse_vio,
                       discard_n_start_poses, discard_n_end_poses) == False:
            has_a_pipeline_failed = True

    # Save boxplots
    if save_boxplots:
        if has_a_pipeline_failed == False:
            print("Saving boxplots.")

            stats = dict()
            for pipeline_type in pipelines_to_run_list:
                results = results_dir + "/" + dataset_name + "/" + pipeline_type + "/results.yaml"
                if not os.path.exists(results):
                    raise Exception("\033[91mCannot plot boxplots: missing results for %s pipeline \
                                    and dataset: %s"%(pipeline_type, dataset_name) + "\033[99m")

                stats[pipeline_type]  = yaml.load(open(results,'r'))
                print("Check stats %s "%(pipeline_type) + results)
                checkStats(stats[pipeline_type])

            print("Drawing boxplots.")
            draw_rpe_boxplots(results_dir + "/" + dataset_name, stats, len(SEGMENTS))
        else:
            print("A pipeline run has failed... skipping boxplot drawing.")

    if not has_a_pipeline_failed:
        print ("All pipeline runs were successful.")
    else:
        print("A pipeline has failed!")
    print("Finished evaluation for dataset: " + dataset_name)
    return has_a_pipeline_failed

def write_flags_parameters(param_name, param_new_value, params_path):
    directory = os.path.dirname(params_path)
    if not os.path.exists(directory):
        raise Exception("\033[91mCould not find directory: " + directory + "\033[99m")
    params_flagfile = open(params_path, "a+")
    params_flagfile.write("--" + param_name + "=" + param_new_value)
    params_flagfile.close()

def check_and_create_regression_test_structure(regression_tests_path, param_names, param_values,
                                               dataset_names, pipeline_types, extra_params_to_modify):
    """ Makes/Checks that the file structure is the correct one, and updates the parameters with the given values"""
    # Make or check regression_test directory
    assert(ensure_dir(regression_tests_path))
    # Make or check param_name directory
    # Use as param_name the concatenated elements of param_names
    param_names_dir = ""
    for i in param_names:
        param_names_dir += str(i) + "-"
    param_names_dir = param_names_dir[:-1]
    assert(ensure_dir("{}/{}".format(regression_tests_path, param_names_dir)))
    for param_value in param_values:
        # Create/Check param_value folder
        param_value_dir = ""
        if isinstance(param_value, list):
            for i in param_value:
                param_value_dir += str(i) + "-"
            param_value_dir = param_value_dir[:-1]
        else:
            param_value_dir = param_value
        ensure_dir("{}/{}/{}".format(regression_tests_path, param_names_dir, param_value_dir))
        # Create params folder by copying from current official one.
        param_dir = "{}/{}/{}/params".format(regression_tests_path, param_names_dir, param_value_dir);
        if (os.path.exists(param_dir)):
            rmtree(param_dir)
        copytree("/home/tonirv/code/evo/results/params", param_dir)

        # Modify param with param value
        for pipeline_type in pipeline_types:
            param_pipeline_dir = "{}/{}".format(param_dir, pipeline_type)
            ensure_dir(param_pipeline_dir)
            written_extra_param_names = []
            is_param_name_written = [False] * len(param_names)
            # VIO params
            vio_file = param_pipeline_dir + "/vioParameters.yaml"
            vio_params = []
            with open(vio_file, 'r') as infile:
                # Skip first yaml line: it contains %YAML:... which can't be read...
                _ = infile.readline()
                vio_params = yaml.load(infile)
                for idx, param_name in enumerate(param_names):
                    if param_name in vio_params:
                        # Modify param_name with param_value
                        if isinstance(param_value, list):
                            vio_params[param_name] = param_value[idx]
                        else:
                            vio_params[param_name] = param_value
                        is_param_name_written[idx] = True
                for extra_param_name, extra_param_value in extra_params_to_modify.iteritems():
                    if extra_param_name in vio_params:
                        vio_params[extra_param_name] = extra_param_value
                        written_extra_param_names.append(extra_param_name)
                # Store param_names with param_value
                with open(vio_file,'w') as outfile:
                    outfile.write("%YAML:1.0\n")
                with open(vio_file,'a') as outfile:
                    outfile.write(yaml.dump(vio_params))

            # Tracker params
            tracker_file = param_pipeline_dir + "/trackerParameters.yaml"
            tracker_params = []
            with open(tracker_file, 'r') as infile:
                # Skip first yaml line: it contains %YAML:... which can't be read...
                _ = infile.readline()
                tracker_params = yaml.load(infile)
                for idx, param_name in enumerate(param_names):
                    if param_name in tracker_params:
                        # Modify param_name with param_value
                        if isinstance(param_value, list):
                            tracker_params[param_name] = param_value[idx]
                        else:
                            tracker_params[param_name] = param_value
                        is_param_name_written[idx] = True
                for extra_param_name, extra_param_value in extra_params_to_modify.iteritems():
                    if extra_param_name in tracker_params:
                        tracker_params[extra_param_name] = extra_param_value
                        written_extra_param_names.append(extra_param_name)
                with open(tracker_file,'w') as outfile:
                    outfile.write("%YAML:1.0\n")
                with open(tracker_file,'a') as outfile:
                    outfile.write(yaml.dump(tracker_params, default_flow_style=False))

            # Gflags
            for idx, param_name in enumerate(param_names):
                if not is_param_name_written[idx]:
                    # Could not find param_name in vio_params nor tracker_params
                    # it must be a gflag:
                    if isinstance(param_value, list):
                        write_flags_parameters(param_name, param_value[idx],
                                               param_pipeline_dir + "/flags/override.flags")
                    else:
                        write_flags_parameters(param_name, param_value,
                                               param_pipeline_dir + "/flags/override.flags")
            for extra_param_name, extra_param_value in extra_params_to_modify.iteritems():
                if extra_param_name not in written_extra_param_names:
                    write_flags_parameters(extra_param_name,
                                           extra_param_value,
                                           param_pipeline_dir + "/flags/override.flags")

        # Create/Check tmp_output folder
        ensure_dir("{}/{}/{}/tmp_output/output".format(regression_tests_path, param_names_dir, param_value_dir))

        for dataset_name in dataset_names:
            ensure_dir("{}/{}/{}/{}".format(regression_tests_path, param_names_dir, param_value_dir, dataset_name))
            # Create ground truth trajectory by copying from current official one.
            copy2("/home/tonirv/code/evo/results/{}/traj_gt.csv".format(dataset_name),
                 "{}/{}/{}/{}/traj_gt.csv".format(regression_tests_path, param_names_dir,
                                                  param_value_dir, dataset_name))
            # Create segments by copying from current official one.
            copy2("/home/tonirv/code/evo/results/{}/segments.txt".format(dataset_name),
                 "{}/{}/{}/{}/segments.txt".format(regression_tests_path, param_names_dir,
                                                  param_value_dir, dataset_name))
            for pipeline_type in pipeline_types:
                ensure_dir("{}/{}/{}/{}/{}".format(regression_tests_path, param_names_dir, param_value_dir,
                                                   dataset_name, pipeline_type))

    # Make/Check results dir for current param_names
    ensure_dir("{}/{}/results".format(regression_tests_path, param_names_dir))
    for dataset_name in dataset_names:
        # Make/Check dataset dir for current param_names_dir, as the performance given the param depends on the dataset.
        ensure_dir("{}/{}/results/{}".format(regression_tests_path, param_names_dir, dataset_name))

def build_list_of_pipelines_to_run(pipelines_to_run):
    pipelines_to_run_list = []
    if pipelines_to_run == 0:
        pipelines_to_run_list = ['S', 'SP', 'SPR']
    if pipelines_to_run == 1:
        pipelines_to_run_list = ['S']
    if pipelines_to_run == 2:
        pipelines_to_run_list = ['SP']
    if pipelines_to_run == 3:
        pipelines_to_run_list = ['SPR']
    if pipelines_to_run == 4:
        pipelines_to_run_list = ['S', 'SP']
    if pipelines_to_run == 5:
        pipelines_to_run_list = ['S', 'SPR']
    if pipelines_to_run == 6:
        pipelines_to_run_list = ['SP', 'SPR']
    return pipelines_to_run_list

def regression_test_simple(test_name, param_names, param_values, only_compile_regression_test_results,
                           run_pipelines, pipelines_to_run, extra_params_to_modify):
    """ Runs the vio pipeline with different values for the given param
    and draws graphs to decide best value for the param:
        - param_names: names of the parameters to fine-tune: e.g ["monoNoiseSigma", "stereoNoiseSigma"]
        - param_values: values that the parameter should take: e.g [[1.0, 1.3], [1.0, 1.2]]
        - only_compile_regression_test_results: just draw boxplots for regression test,
            skip all per pipeline analysis and runs, assumes we have results.yaml for
            each param value, dataset and pipeline.
        - run_pipelines: run pipelines, if set to false, it won't run pipelines and will assume we have a traj_est.csv already.
        - pipelines_to_run: which pipeline to run, useful when a parameter only affects a single pipeline."""
    # Ensure input is correct.
    if isinstance(param_names, list):
        if len(param_names) > 1:
            assert(len(param_names) == len(param_values[0]))
            for i in range(2, len(param_names)):
                # Ensure all rows have the same number of parameter changes
                assert(len(param_values[i-2]) == len(param_values[i-1]))

    # Check and create file structure
    dataset_names = ["V1_01_easy"]
    pipelines_to_run_list = build_list_of_pipelines_to_run(pipelines_to_run);
    REGRESSION_TESTS_DIR = "/home/tonirv/code/evo-1/regression_tests/" + test_name
    check_and_create_regression_test_structure(REGRESSION_TESTS_DIR, param_names, param_values,
                                               dataset_names, pipelines_to_run_list, extra_params_to_modify)

    param_names_dir = ""
    for i in param_names:
        param_names_dir += str(i) + "-"
    param_names_dir = param_names_dir[:-1]
    DATASET_DIR = '/home/tonirv/datasets/EuRoC'
    BUILD_DIR = '/home/tonirv/code/spark_vio/build'
    if not only_compile_regression_test_results:
        for param_value in param_values:
            param_value_dir = ""
            if isinstance(param_value, list):
                for i in param_value:
                    param_value_dir += str(i) + "-"
                param_value_dir = param_value_dir[:-1]
            else:
                param_value_dir = param_value
            results_dir = "{}/{}/{}".format(REGRESSION_TESTS_DIR, param_names_dir, param_value_dir)
            pipeline_output_dir = results_dir + "/tmp_output/output"
            for dataset_name in dataset_names:
                run_dataset(results_dir, DATASET_DIR, dataset_name, BUILD_DIR,
                               run_pipelines, # Should we re-run pipelines?
                               True, # Should we run the analysis of per pipeline errors?
                               False, # Should we display plots?
                               True, # Should we save results?
                               True, # Should we save plots?
                               False, # Should we save boxplots?
                               pipelines_to_run_list) # Should we run 0: all pipelines, 1: S, 2:SP 3:SPR

                print("Finished analysis of pipelines for param_value: {} for parameter: {}".format(param_value_dir, param_names_dir))
                print("Finished pipeline runs/analysis for regression test of param_name: {}".format(param_names_dir))

    # Compile results for current param_name
    print("Drawing boxplot APE for regression test of param_name: {}".format(param_names_dir))
    for dataset_name in dataset_names:
        stats = dict()
        for param_value in param_values:
            param_value_dir = ""
            if isinstance(param_value, list):
                for i in param_value:
                    param_value_dir += str(i) + "-"
                param_value_dir = param_value_dir[:-1]
            else:
                param_value_dir = param_value
            stats[param_value_dir] = dict()
            for pipeline in pipelines_to_run_list:
                results_file = "{}/{}/{}/{}/{}/results.yaml".format(REGRESSION_TESTS_DIR, param_names_dir,
                                                                    param_value_dir, dataset_name, pipeline)
                if os.path.isfile(results_file):
                    stats[param_value_dir][pipeline] = yaml.load(open(results_file,'r'))
                else:
                    print("Could not find results file: {}".format(results_file) + ". Adding cross to boxplot...")
                    stats[param_value_dir][pipeline] = False

        print("Drawing regression simple APE boxplots for dataset: " + dataset_name)
        plot_dir = "{}/{}/results/{}".format(REGRESSION_TESTS_DIR, param_names_dir, dataset_name)
        max_y = -1
        if dataset_name == "V2_02_medium":
            max_y = 0.40
        if dataset_name == "V1_01_easy":
            max_y = 0.20
        draw_regression_simple_boxplot_APE(param_names, stats, plot_dir, max_y)
    print("Finished regression test for param_name: {}".format(param_names_dir))

def run(args):
    import evo.common_ape_rpe as common
    from evo.tools import log
    from evo.tools.settings import SETTINGS

    log.configure_logging(args.verbose, args.silent, args.debug)
    if args.debug:
        from pprint import pformat
        parser_str = pformat({arg: getattr(args, arg) for arg in vars(args)})
        logger.debug("main_parser config:\n{}".format(parser_str))
        logger.debug(SEP)

    RESULTS_DIR = '/home/tonirv/code/evo-1/results'
    DATASET_DIR = '/home/tonirv/datasets/EuRoC'
    BUILD_DIR = '/home/tonirv/code/spark_vio/build'

    # Comment out any experiment that you do not want to run
    LIST_OF_EXPERIMENTS_TO_RUN = [\
                                  # 'MH_01_easy',
                                  # 'MH_02_easy',
                                  # 'MH_03_medium',
                                  # 'mh_04_difficult', # Diff number of left/right imgs...
                                  # 'MH_05_difficult',
                                  'V1_01_easy',
                                  # 'V1_02_medium',
                                  # 'V1_03_difficult',
                                  # 'V2_01_easy',
                                  # 'V2_02_medium',
                                  # 'v2_03_difficult' # Diff number of left/right imgs...
                                 ]

    # Run experiments.
    print("Run experiments")
    for dataset_name in LIST_OF_EXPERIMENTS_TO_RUN:
        print("Run dataset:%s", dataset_name)
        pipelines_to_run_list = build_list_of_pipelines_to_run(args.pipeline_type)
        run_dataset(RESULTS_DIR, DATASET_DIR, dataset_name, BUILD_DIR,
                    args.run_pipeline, args.analyse_vio,
                    args.plot, args.save_results,
                    args.save_plots, args.save_boxplots, pipelines_to_run_list,
                    args.discard_n_start_poses, args.discard_n_end_poses)

def parser():
    import argparse
    basic_desc = "Full evaluation of pipeline (APE trans + RPE trans + RPE rot) metric app"

    shared_parser = argparse.ArgumentParser(add_help=True, description="{}".format(basic_desc))
    algo_opts = shared_parser.add_argument_group("algorithm options")
    output_opts = shared_parser.add_argument_group("output options")
    usability_opts = shared_parser.add_argument_group("usability options")

    algo_opts.add_argument("-p", "--pipeline_type", type=float,
                           help="Pipeline_type = 0:All 1:S 2:SP 3:SPR -1:None", default = 0)
    algo_opts.add_argument("-s", "--discard_n_start_poses", type=float,
                           help="Discard n start poses from ground-truth trajectory.", default = 0)
    algo_opts.add_argument("-e", "--discard_n_end_poses", type=float,
                           help="Discard n end poses from ground-truth trajectory", default = 0)
    algo_opts.add_argument("-r", "--run_pipeline", action="store_true",
                           help="Run vio?")
    algo_opts.add_argument("-a", "--analyse_vio", action="store_true",
                           help="Analyse vio, compute APE and RPE")

    output_opts.add_argument("--plot", action="store_true", help="show plot window",)
    output_opts.add_argument("--plot_mode", default="xyz", help="the axes for plot projection",
                             choices=["xy", "yx", "xz", "zx", "yz", "xyz"])
    output_opts.add_argument("--plot_colormap_max", type=float,
                             help="The upper bound used for the color map plot "
                             "(default: maximum error value)")
    output_opts.add_argument("--plot_colormap_min", type=float,
                             help="The lower bound used for the color map plot "
                             "(default: minimum error value)")
    output_opts.add_argument("--plot_colormap_max_percentile", type=float,
                             help="Percentile of the error distribution to be used "
                             "as the upper bound of the color map plot "
                             "(in %%, overrides --plot_colormap_min)")
    output_opts.add_argument("--save_plots", action="store_true",
                             help="Save plots?")
    output_opts.add_argument("--save_boxplots", action="store_true",
                             help="Save boxplots?")
    output_opts.add_argument("--save_results", action="store_true",
                             help="Save results?")

    usability_opts.add_argument("--no_warnings", action="store_true",
                                help="no warnings requiring user confirmation")
    usability_opts.add_argument("-v", "--verbose", action="store_true",
                                help="verbose output")
    usability_opts.add_argument("--silent", action="store_true",
                                help="don't print any output")
    usability_opts.add_argument("--debug", action="store_true",
                                help="verbose output with additional debug info")
    usability_opts.add_argument("-c", "--config",
                                help=".json file with parameters (priority over command line args)")

    main_parser = argparse.ArgumentParser(
        description="{}".format(basic_desc))
    sub_parsers = main_parser.add_subparsers(dest="subcommand")
    sub_parsers.required = True

    return shared_parser

if __name__ == '__main__':
    from evo import entry_points
    entry_points.evaluation()



# -*- coding: UTF8 -*-
"""
cross-compatible functions as a workaround until we can drop Python 2.7
author: Michael Grupp

This file is part of evo (github.com/MichaelGrupp/evo).

evo is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

evo is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with evo.  If not, see <http://www.gnu.org/licenses/>.
"""

import os
import sys


def which(executable):
    """
    behaves like shutil.which() from Python 3.3+
    :param executable: string
    :return: path of the executable or None
    """
    if sys.version_info >= (3, 3):
        import shutil
        return shutil.which(executable)
    else:
        from distutils.spawn import find_executable
        path = find_executable(executable)
        if path is not None:
            return path if os.access(path, os.X_OK) else None
        else:
            return path



# -*- coding: UTF8 -*-
"""
utilities for the configuration of the package's loggers
author: Michael Grupp

This file is part of evo (github.com/MichaelGrupp/evo).

evo is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

evo is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with evo.  If not, see <http://www.gnu.org/licenses/>.
"""

import logging
import os
import sys

import colorama
from colorama import Fore

from evo.tools.settings import SETTINGS, DEFAULT_LOGFILE_PATH

CONSOLE_ERROR_FMT = "{}[%(levelname)s]{} %(msg)s".format(
    Fore.LIGHTRED_EX, Fore.RESET)
CONSOLE_WARN_FMT = "{}[%(levelname)s]{} %(msg)s".format(
    Fore.LIGHTYELLOW_EX, Fore.RESET)
DEFAULT_LONG_FMT = "[%(levelname)s][%(asctime)s][%(module)s.%(funcName)s():%(lineno)s]\n%(message)s"


class ConsoleFormatter(logging.Formatter):
    def __init__(self, fmt="%(msg)s"):
        logging.Formatter.__init__(self, fmt)
        colorama.init()
        self.critical_fmt = CONSOLE_ERROR_FMT
        self.error_fmt = CONSOLE_ERROR_FMT
        self.warning_fmt = CONSOLE_WARN_FMT
        self.info_fmt = fmt
        self.debug_fmt = fmt

    def format(self, record):
        if record.levelno == logging.CRITICAL:
            self._fmt = self.error_fmt
        elif record.levelno == logging.ERROR:
            self._fmt = self.error_fmt
        elif record.levelno == logging.WARNING:
            self._fmt = self.warning_fmt
        elif record.levelno == logging.INFO:
            self._fmt = self.info_fmt
        elif record.levelno == logging.DEBUG:
            self._fmt = self.debug_fmt
        result = logging.Formatter.format(self, record)
        return result


# configures the package's root logger (see __init__.py)
def configure_logging(verbose=False, silent=False, debug=False,
                      console_fmt=None, file_fmt=DEFAULT_LONG_FMT,
                      file_path=None):

    logger = logging.getLogger("evo")
    logger.setLevel(logging.DEBUG)
    logger.propagate = False

    if len(logger.handlers) > 0:
        logger.removeHandler(logger.handlers[0])

    if not SETTINGS.logfile_enabled:
        logfile = os.devnull
    elif file_path is None:
        logfile = DEFAULT_LOGFILE_PATH
    else:
        logfile = file_path

    file_handler = logging.FileHandler(logfile)
    file_handler.setLevel(logging.DEBUG)
    file_handler.setFormatter(logging.Formatter(file_fmt))
    logger.addHandler(file_handler)

    if debug or verbose:
        console_level = logging.DEBUG
    elif silent:
        console_level = logging.WARNING
    else:
        console_level = logging.INFO

    if debug:
        console_fmt = DEFAULT_LONG_FMT
    elif console_fmt is None:
        console_fmt = SETTINGS.logging_format

    console_handler = logging.StreamHandler(stream=sys.stdout)
    console_handler.setLevel(console_level)
    console_handler.setFormatter(ConsoleFormatter(console_fmt))
    logger.addHandler(console_handler)

    # log header for debug mode
    if debug:
        import getpass as gp
        import platform as pf
        logger.debug(
            "System info:\nPython {pyversion}\n{platform}\n{user}\n".format(
                pyversion=pf.python_version(), platform=pf.platform(),
                user=gp.getuser() + "@" + pf.node()))



"""
filter algorithms
author: Michael Grupp

This file is part of evo (github.com/MichaelGrupp/evo).

evo is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

evo is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with evo.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import print_function

import sys
import logging

import numpy as np

from evo.core import geometry, trajectory
from evo.core import lie_algebra as lie

logger = logging.getLogger(__name__)


class FilterException(Exception):
    pass


def bounded_binary_search(generator, length, target, lower_bound, upper_bound):
    """
    efficient binary search for a <target> value within bounds
    [<lower_bound>, <upper_bound>]
    - converges to a locally optimal result within the bounds
    - instead of indexing an iterable, lazy evaluate a functor for performance
    :param generator: a generator or functor that yields a value of the search
                      area given an index
    :param length: full length of the search area
    :param target: the value to search
    :param lower_bound: the lower bound up to which results are accepted
    :param upper_bound: the upper bound up to which results are accepted
    :return: success: (True, the index of the target) - fail: (False, -1)
    """
    start, mid = 0, -1
    end = length - 1
    residual = 0.0
    found = False
    num_iter = 0
    while start <= end and not found:
        num_iter += 1
        mid = (start + end) // 2
        val = generator(mid)
        if lower_bound <= val <= upper_bound:
            residual = abs(val - target)
            if abs(generator(mid - 1) - target) <= residual:
                end = mid - 1
                continue  # refinement possible in left direction
            elif abs(generator(mid + 1) - target) < residual:
                start = mid + 1
                continue  # refinement possible in right direction
            else:
                found = True  # converged
        if not found:
            if target < val:
                end = mid - 1
            else:
                start = mid + 1
    return found, mid, residual, num_iter


def filter_pairs_by_index(poses, delta, all_pairs=False):
    """
    filters pairs in a list of SE(3) poses by their index distance
    :param poses: list of SE(3) poses
    :param delta: the index distance used for filtering
    :param all_pairs: use all pairs instead of consecutive pairs
    :return: list of index tuples of the filtered pairs
    """
    if all_pairs:
        ids = range(len(poses))
        id_pairs = [(i, i + delta) for i in ids if i + delta < len(poses)]
    else:
        ids = np.arange(0, len(poses), delta)
        id_pairs = [(i, j) for i, j in zip(ids, ids[1:])]
    return id_pairs


def filter_pairs_by_distance(poses, delta, tol=0.0, all_pairs=False):
    """
    filters pairs in a list of SE(3) poses by their direct distance in meters
     - only the direct distance between the two pair points is considered
    :param poses: list of SE(3) poses
    :param delta: the distance in meters used for filtering
    :param tol: absolute distance tolerance to accept or reject pairs
                in all_pairs mode
    :param all_pairs: use all pairs instead of consecutive pairs
    :return: list of index tuples of the filtered pairs
    """
    if all_pairs:
        upper_bound = delta + tol
        lower_bound = delta - tol
        id_pairs = []
        ids = range(len(poses))
        positions = [pose[:3, 3] for pose in poses]
        for i in ids:
            for j in ids[i + 1:]:
                current_dist = abs(np.linalg.norm(positions[i] - positions[j]))
                if lower_bound <= current_dist <= upper_bound:
                    id_pairs.append((i, j))
    else:
        ids = []
        i = 0
        while i < len(poses):
            current_pose = poses[i]
            for j, next_pose in enumerate(poses[i:]):
                j += i
                current_dist = np.linalg.norm(current_pose[:3, 3] -
                                              next_pose[:3, 3])
                if current_dist >= delta:
                    ids.append(i)
                    i = j
                    break
            i += 1
        id_pairs = [(i, j) for i, j in zip(ids, ids[1:])]
    return id_pairs


def filter_pairs_by_path(poses, delta, tol=0.0, all_pairs=False):
    """
    filters pairs in a list of SE(3) poses by their path distance in meters
     - the accumulated, traveled path distance between the two pair points
       is considered
    :param poses: list of SE(3) poses
    :param delta: the path distance in meters used for filtering
    :param tol: absolute path tolerance to accept or reject pairs
                in all_pairs mode
    :param all_pairs: use all pairs instead of consecutive pairs
    :return: list of index tuples of the filtered pairs
    """
    if all_pairs:
        upper_bound = delta + tol
        lower_bound = delta - tol
        id_pairs = []
        ids = range(len(poses))
        positions = np.array([pose[:3, 3] for pose in poses])
        res_avg = 0.0
        n_iter_avg = 0.0
        print_progress = logger.isEnabledFor(logging.DEBUG)
        num_pairs = 0
        for i in ids:
            found, j, res, n = bounded_binary_search(
                lambda x: geometry.arc_len(positions[i:x + 1]), len(positions),
                delta, lower_bound, upper_bound)
            n_iter_avg += n
            if found:
                num_pairs += 1
                res_avg += res
                id_pairs.append((i, j))
            if print_progress:
                print("\rSearching", delta, "m path sub-sequences - found",
                      num_pairs, end="\r")
                sys.stdout.flush()
        if print_progress:
            print("")
        if num_pairs != 0:
            logger.debug("avg. target residual: " +
                         "{0:.6f}".format(res_avg / num_pairs) + "m" +
                         " | avg. num. iterations: " +
                         "{0:.6f}".format(n_iter_avg / num_pairs))
        else:
            logger.debug("Found no pairs for delta '" + str(delta) + "m'.")
    else:
        ids = []
        previous_pose = poses[0]
        current_path = 0.0
        for i, current_pose in enumerate(poses):
            current_path += np.linalg.norm(current_pose[:3, 3] -
                                           previous_pose[:3, 3])
            previous_pose = current_pose
            if current_path >= delta:
                ids.append(i)
                current_path = 0.0
        id_pairs = [(i, j) for i, j in zip(ids, ids[1:])]
    return id_pairs


def filter_pairs_by_angle(poses, delta, tol=0.0, degrees=False,
                          all_pairs=False):
    """
    filters pairs in a list of SE(3) poses by their absolute relative angle
     - by default, the angle accumulated on the path between the two pair poses
       is considered
     - if <all_pairs> is set to True, the direct angle between the two pair
       poses is considered
    :param poses: list of SE(3) poses
    :param delta: the angle in radians used for filtering
    :param tol: absolute angle tolerance to accept or reject pairs
                in all_pairs mode
    :param degrees: set to True if <delta> is in degrees instead of radians
    :param all_pairs: use all pairs instead of consecutive pairs
    :return: list of index tuples of the filtered pairs
    """
    if all_pairs:
        upper_bound = delta + tol
        lower_bound = delta - tol
        id_pairs = []
        ids = range(len(poses))
        if degrees:
            angles = [lie.so3_log(p[:3, :3]) * 180 / np.pi for p in poses]
        else:
            angles = [lie.so3_log(p[:3, :3]) for p in poses]
        for i in ids:
            for j in ids[i + 1:]:
                current_angle = abs(angles[i] - angles[j])
                if lower_bound <= current_angle <= upper_bound:
                    id_pairs.append((i, j))
    else:
        ids = []
        if degrees:
            angles = [lie.so3_log(p[:3, :3]) * 180 / np.pi for p in poses]
        else:
            angles = [lie.so3_log(p[:3, :3]) for p in poses]
        previous_angle = angles[0]
        current_delta = 0.0
        ids.append(0)
        for i, current_angle in enumerate(angles):
            current_delta += abs(current_angle - previous_angle)
            previous_angle = current_angle
            if current_delta >= delta:
                ids.append(i)
                current_delta = 0.0
        id_pairs = [(i, j) for i, j in zip(ids, ids[1:])]
    return id_pairs


def filter_pairs_by_speed(poses, timestamps, speed, tol):
    """
    filters pairs in a list of SE(3) poses by the linear speed of the motion
    in between them
    :param poses: list of SE(3) poses
    :param timestamps: list of timestamps corresponding to the poses
    :param speed: in m/s
    :param tol: tolerance to accept or reject velocities, in m/s
    :return: list of index tuples of the filtered pairs
    """
    positions = [pose[:3, 3] for pose in poses]
    speeds = [
        trajectory.calc_speed(positions[i], positions[i + 1], timestamps[i],
                              timestamps[i + 1])
        for i in range(len(positions) - 1)
    ]
    id_pairs = [(i, i + 1) for i, v in enumerate(speeds)
                if speed - tol <= v <= speed + tol]
    return id_pairs


def filter_pairs_by_angular_speed(poses, timestamps, speed, tol,
                                  degrees=False):
    """
    filters pairs in a list of SE(3) poses by the angular speed of the motion
    in between them
    :param poses: list of SE(3) poses
    :param timestamps: list of timestamps corresponding to the poses
    :param speed: in rad/s
    :param tol: tolerance to accept or reject velocities, in rad/s
    :param degrees: set to True to use deg/s
    :return: list of index tuples of the filtered pairs
    """
    speeds = [
        trajectory.calc_angular_speed(poses[i], poses[i + 1], timestamps[i],
                                      timestamps[i + 1], degrees)
        for i in range(len(poses) - 1)
    ]
    id_pairs = [(i, i + 1) for i, v in enumerate(speeds)
                if speed - tol <= v <= speed + tol]
    return id_pairs


def id_pairs_from_delta(poses, delta, delta_unit, rel_tol=0.1,
                        all_pairs=False):
    """
    high-level function - get index tuples of pairs with distance==delta
    from a pose list
    :param poses: list of SE(3) poses
    :param delta: the interval step for indices
    :param delta_unit: unit of delta (metrics.Unit enum member)
    :param rel_tol: relative tolerance to accept or reject deltas
    :param all_pairs: use all pairs instead of consecutive pairs
    :return: list of index tuples (pairs)
    """
    from evo.core.metrics import Unit
    if delta_unit == Unit.frames:
        id_pairs = filter_pairs_by_index(poses, delta, all_pairs)
    elif delta_unit == Unit.meters:
        id_pairs = filter_pairs_by_path(poses, delta, delta * rel_tol,
                                        all_pairs)
    elif delta_unit in {Unit.degrees, Unit.radians}:
        use_degrees = (delta_unit == Unit.degrees)
        id_pairs = filter_pairs_by_angle(poses, delta, delta * rel_tol,
                                         use_degrees, all_pairs)
    else:
        raise FilterException("unsupported delta unit: {}".format(delta_unit))

    if len(id_pairs) == 0:
        raise FilterException(
            "delta = {} ({}) produced an empty index list - try lower values"
            "or a less strict tolerance".format(delta, delta_unit.value))

    logger.debug(
        "Found {} pairs with delta {} ({}) "
        "among {} poses ".format(
            len(id_pairs), delta, delta_unit.value, len(poses)) +
        ("using consecutive pairs." if not all_pairs else "using all pairs."))

    return id_pairs



# -*- coding: UTF8 -*-
"""
translate between evo and Pandas types
author: Michael Grupp

This file is part of evo (github.com/MichaelGrupp/evo).

evo is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

evo is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with evo.  If not, see <http://www.gnu.org/licenses/>.
"""

import os

import numpy as np
import pandas as pd

from evo.core import trajectory, result


def trajectory_to_df(traj):
    if not isinstance(traj, trajectory.PosePath3D):
        raise TypeError("trajectory.PosePath3D or derived required")
    poses_dict = {
        "x": traj.positions_xyz[:, 0],
        "y": traj.positions_xyz[:, 1],
        "z": traj.positions_xyz[:, 2],
        "qw": traj.orientations_quat_wxyz[:, 0],
        "qx": traj.orientations_quat_wxyz[:, 1],
        "qy": traj.orientations_quat_wxyz[:, 2],
        "qz": traj.orientations_quat_wxyz[:, 3],
    }
    if type(traj) is trajectory.PoseTrajectory3D:
        index = traj.timestamps
    else:
        index = np.arange(0, traj.num_poses)
    return pd.DataFrame(data=poses_dict, index=index)


def result_to_df(result_obj, label=None):
    if not isinstance(result_obj, result.Result):
        raise TypeError("result.Result or derived required")
    data = {
        "info": result_obj.info,
        "stats": result_obj.stats,
        "np_arrays": {},
        "trajectories": {}
    }
    for name, array in result_obj.np_arrays.items():
        data["np_arrays"][name] = array
    if label is None and "est_name" in data["info"]:
        label = os.path.splitext(os.path.basename(data["info"]["est_name"]))[0]
    elif label is None:
        label = "unnamed_result"
    return pd.DataFrame(data=data).T.stack().to_frame(name=label)



#!/usr/bin/env python
"""
unit test for lie_algebra module - mainly test mathematical correctness
author: Michael Grupp

This file is part of evo (github.com/MichaelGrupp/evo).

evo is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

evo is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with evo.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import print_function

import random
import timeit
import unittest

import numpy as np

import context
from evo.core import lie_algebra as lie


class TestSE3(unittest.TestCase):
    def test_is_se3(self):
        # yapf: disable
        p = np.array([[1, 0, 0, 1],
                      [0, 0, -1, 2],
                      [0, 1, 0, 3],
                      [0, 0, 0, 1]])
        self.assertTrue(lie.is_se3(p))
        p_false = np.array([[1, 0, 0, 1],
                            [0, 0, -4223, 2],
                            [0, 111, 0, 3],
                            [0, 0, 0, 1]])
        self.assertFalse(lie.is_se3(p_false))
        # yapf: enable

    def test_random_se3(self):
        self.assertTrue(lie.is_se3(lie.random_se3()))

    def test_se3_inverse(self):
        p = lie.random_se3()
        p_inv = lie.se3_inverse(p)
        self.assertTrue(lie.is_se3(p_inv))
        self.assertTrue(np.allclose(p_inv.dot(p), np.eye(4)))

    def test_relative_se3(self):
        a = lie.random_se3()
        b = lie.random_se3()
        self.assertTrue(lie.is_se3(a) and lie.is_se3(b))
        a_to_b = lie.relative_se3(a, b)
        self.assertTrue(lie.is_se3(a_to_b))
        b_from_a = a.dot(a_to_b)
        self.assertTrue(np.allclose(b_from_a, b))


class TestSO3(unittest.TestCase):
    def test_is_so3(self):
        # yapf: disable
        r = np.array([[1, 0, 0],
                      [0, 0, -1],
                      [0, 1, 0]])
        self.assertTrue(lie.is_so3(r))
        # yapf: enable

    def test_random_so3(self):
        r = lie.random_so3()
        self.assertTrue(lie.is_so3(r))

    def test_relative_so3(self):
        a = lie.random_so3()
        b = lie.random_so3()
        self.assertTrue(lie.is_so3(a) and lie.is_so3(b))
        a_to_b = lie.relative_so3(a, b)
        b_from_a = a.dot(a_to_b)
        self.assertTrue(np.allclose(b_from_a, b))

    def test_so3_from_se3(self):
        p = lie.random_se3()
        r = lie.so3_from_se3(p)
        self.assertTrue(lie.is_so3(r))

    def test_so3_log_exp(self):
        r = lie.random_so3()
        self.assertTrue(lie.is_so3(r))
        axis, angle = lie.so3_log(r, return_angle_only=False)
        self.assertTrue(np.allclose(r, lie.so3_exp(axis, angle), atol=1e-6))
        angle = lie.so3_log(r)
        # we ignore signs here, therefore check also transpose
        self.assertTrue(
            np.allclose(r,
                        lie.so3_exp(axis, angle).T)
            or np.allclose(r, lie.so3_exp(axis, angle)))

    def test_so3_log_exp_skew(self):
        r = lie.random_so3()
        log = lie.so3_log(r, return_skew=True)  # skew-symmetric tangent space
        # here, axis is a rotation vector with norm = angle
        axis = lie.vee(log)
        angle = np.linalg.norm(axis)
        self.assertTrue(np.allclose(r, lie.so3_exp(axis, angle)))


class TestSim3(unittest.TestCase):
    def test_is_sim3(self):
        r = lie.random_so3()
        t = np.array([1, 2, 3])
        s = 3
        p = lie.sim3(r, t, s)
        self.assertTrue(lie.is_sim3(p, s))

    def test_sim3_scale_effect(self):
        r = lie.random_so3()
        t = np.array([0, 0, 0])
        s = random.random() * 10
        x = np.array([1, 0, 0, 1]).T  # homogeneous vector
        p = lie.sim3(r, t, s)
        self.assertTrue(lie.is_sim3(p, s))
        x = p.dot(x)  # apply Sim(3) transformation
        self.assertTrue(
            np.equal(x,
                     lie.se3(r).dot(np.array([s, 0, 0, 1]))).all())


if __name__ == '__main__':
    """
    benchmarks
    """
    print("\ncheck speed of SE(3) inverse:")
    setup = "from evo.core import lie_algebra as lie; " \
            "import numpy as np; se3 = lie.random_se3()"
    print("time for 1000*lie.se3_inverse(se3): ",
          timeit.timeit("lie.se3_inverse(se3)", setup=setup, number=1000))
    print("time for 1000*np.linalg.inv(se3): ",
          timeit.timeit("np.linalg.inv(se3)", setup=setup, number=1000))

    print("\ncheck speed of  SO(3) log:")
    setup = "from evo.core import lie_algebra as lie; " \
            "import numpy as np; so3 = lie.random_so3()"
    print("time for 1000*lie.so3_log(so3, skew=True): ",
          timeit.timeit("lie.so3_log(so3, True)", setup=setup, number=1000))
    print("time for 1000*lie.so3_log(so3): ",
          timeit.timeit("lie.so3_log(so3)", setup=setup, number=1000))
    setup = "from evo.core import lie_algebra as lie; import numpy as np; " \
            "import evo.core.transformations as tr; " \
            "so3 = lie.se3(lie.random_so3(), [0, 0, 0])"
    print("time for 1000*tr.rotation_from_matrix(so3): ",
          timeit.timeit("tr.rotation_from_matrix(so3)", setup=setup,
                        number=1000))
    setup = "from evo.core import lie_algebra as lie; " \
            "import numpy as np; so3 = lie.random_so3(); " \
            "axis, angle = lie.so3_log(so3, False)"
    print("time for 1000*lie.so3_exp(axis, angle): ",
          timeit.timeit("lie.so3_exp(axis, angle)", setup=setup, number=1000))
    """
    unit test
    """
    unittest.main(verbosity=2)



# -*- coding: UTF8 -*-
"""
Provides algorithms for time synchronization.
author: Michael Grupp

This file is part of evo (github.com/MichaelGrupp/evo).

evo is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

evo is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with evo.  If not, see <http://www.gnu.org/licenses/>.
"""

import copy
import logging

import numpy as np

from evo.core.trajectory import PoseTrajectory3D

logger = logging.getLogger(__name__)


class SyncException(Exception):
    pass


def matching_time_indices(stamps_1, stamps_2, max_diff=0.01, offset_2=0.0):
    """
    Searches for the best matching timestamps of two lists of timestamps
    and returns the list indices of the best matches.
    :param stamps_1: first vector of timestamps
    :param stamps_2: second vector of timestamps
    :param max_diff: max. allowed absolute time difference
    :param offset_2: optional offset of second vector
    :return: the indices of the matching stamps in stamps_1
    """
    matching_indices = []
    stamps_2 = copy.deepcopy(stamps_2)
    stamps_2 += offset_2
    for stamp in stamps_1:
        diffs = np.abs(stamps_2 - stamp)
        argmin = np.argmin(diffs)
        if diffs[argmin] <= max_diff:
            matching_indices.append(argmin)
    return matching_indices


def associate_trajectories(traj_1, traj_2, max_diff=0.01, offset_2=0.0,
                           first_name="first trajectory",
                           snd_name="second trajectory"):
    """
    Synchronizes two trajectories by matching their timestamps.
    :param traj_1: trajectory.PoseTrajectory3D object of first trajectory
    :param traj_2: trajectory.PoseTrajectory3D object of second trajectory
    :param max_diff: max. allowed absolute time difference for associating
    :param offset_2: optional time offset of second trajectory
    :param first_name: name of first trajectory for verbose logging
    :param snd_name: name of second trajectory for verbose/debug logging
    :return: traj_1, traj_2 (synchronized)
    """
    if not isinstance(traj_1, PoseTrajectory3D) \
        or not isinstance(traj_2, PoseTrajectory3D):
        raise SyncException("trajectories must be PoseTrajectory3D objects")

    snd_longer = len(traj_2.timestamps) > len(traj_1.timestamps)
    traj_long = copy.deepcopy(traj_2) if snd_longer else copy.deepcopy(traj_1)
    traj_short = copy.deepcopy(traj_1) if snd_longer else copy.deepcopy(traj_2)
    max_pairs = len(traj_long.timestamps)

    # First, match the timestamps of the shorter trajectory to the longer one.
    matching_indices = matching_time_indices(
        traj_short.timestamps, traj_long.timestamps, max_diff,
        offset_2 if snd_longer else -offset_2)
    traj_long.reduce_to_ids(matching_indices)

    # Next, reversely match the reduced long trajectory to the shorter one.
    matching_indices = matching_time_indices(
        traj_long.timestamps, traj_short.timestamps, max_diff,
        -offset_2 if snd_longer else offset_2)
    traj_short.reduce_to_ids(matching_indices)

    traj_1 = traj_short if snd_longer else traj_long
    traj_2 = traj_long if snd_longer else traj_short

    if len(matching_indices) == 0:
        raise SyncException(
            "found no matching timestamps between {} and {} with max. time "
            "diff {} (s) and time offset {} (s)".format(
                first_name, snd_name, max_diff, offset_2))

    logger.debug(
        "Found {} of max. {} possible matching timestamps between...\n"
        "\t{}\nand:\t{}\n..with max. time diff.: {} (s) "
        "and time offset: {} (s).".format(
            len(matching_indices), max_pairs, first_name, snd_name, max_diff,
            offset_2))

    return traj_1, traj_2



#!/usr/bin/env python
# -*- coding: UTF8 -*-
# PYTHON_ARGCOMPLETE_OK
"""
simple JSON configuration generator script for executables
author: Michael Grupp

This file is part of evo (github.com/MichaelGrupp/evo).

evo is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

evo is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with evo.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import print_function

import os
import sys
import json
import argparse
import logging

import colorama
from colorama import Fore, Style
from pygments import highlight, lexers, formatters

from evo.tools import log, user, settings
from evo.tools.settings_template import DEFAULT_SETTINGS_DICT_DOC
from evo.tools.settings_template import DEFAULT_SETTINGS_DICT

logger = logging.getLogger(__name__)

SEP = "-" * 80


class ConfigError(Exception):
    pass


def log_info_dict_json(data_str, colored=True):
    data_str = json.dumps(data_str, indent=4, sort_keys=True)
    if colored and os.name != "nt":
        data_str = highlight(data_str, lexers.JsonLexer(),
                             formatters.Terminal256Formatter(style="monokai"))
    logger.info(data_str)


def show(cfg_path, colored=True):
    with open(cfg_path) as cfg_file:
        log_info_dict_json(json.load(cfg_file), colored)


def merge_json_union(first, second, soft=False):
    with open(first, 'r+') as f_1:
        cfg_1 = json.loads(f_1.read())
        with open(second) as f_2:
            cfg_2 = json.loads(f_2.read())
            cfg_1 = settings.merge_dicts(cfg_1, cfg_2, soft)
        f_1.truncate(0)
        f_1.seek(0)
        f_1.write(json.dumps(cfg_1, indent=4, sort_keys=True))


def is_number(s):
    try:
        float(s)
        return True
    except ValueError:
        return False


def set_cfg(cfg_path, arg_list):
    with open(cfg_path) as cfg_file:
        data = json.load(cfg_file)
    max_idx = len(arg_list) - 1
    for i, arg in enumerate(arg_list):
        if arg in data:
            if i + 1 <= max_idx:
                if arg_list[i + 1].lower() == "true":
                    data[arg] = True
                elif arg_list[i + 1].lower() == "false":
                    data[arg] = False
                else:
                    values = []
                    for j in range(i + 1, max_idx + 1):
                        value = arg_list[j]
                        if value in data:
                            break
                        if is_number(value):
                            if int(float(value)) - float(value) != 0:
                                values.append(float(value))
                            else:
                                values.append(int(float(value)))
                        else:
                            values.append(value)
                    if len(values) == 1:
                        values = values[0]
                    data[arg] = not data[arg] if isinstance(data[arg],
                                                            bool) else values
            elif i + 1 > max_idx or arg_list[i + 1] in data:
                # toggle boolean parameter
                data[arg] = not data[arg] if isinstance(data[arg],
                                                        bool) else data[arg]
    with open(cfg_path, 'w') as cfg_file:
        cfg_file.write(json.dumps(data, indent=4, sort_keys=True))


def generate(arg_list):
    data = {}
    max_idx = len(arg_list) - 1
    for i, arg in enumerate(arg_list):
        if arg.startswith("-"):
            arg = arg[1:] if not arg.startswith("--") else arg[2:]
            if (i + 1 <= max_idx
                    and arg_list[i + 1].startswith("-")) or i + 1 > max_idx:
                data[arg] = True  # just a boolean flag
            else:
                values = []
                for j in range(i + 1, max_idx + 1):
                    value = arg_list[j]
                    if value.startswith("-"):
                        break
                    values.append(float(value) if is_number(value) else value)
                if len(values) == 1:
                    values = values[0]
                data[arg] = values
    return data


SET_HELP = '''
set parameters
Unless -c / --config is specified, the package settings will be used.

--EXAMPLE--

If your configuration looks like this (via 'evo_config show'):

    {
        "plot_export_format": "svg"
        "plot_info_text": true,
    }
    
running:
    evo_config set plot_export_format png plot_info_text
will set it to:

    {
        "plot_export_format": "png"
        "plot_info_text": false,
    }
'''

GENERATE_HELP = '''
generate configuration files from command-line args

The configuration files are intended to hold command line
parameters used by the respective executables, e.g. 'evo_ape'.

--EXAMPLE--

Running:

    evo_config generate --align --plot --plot_mode xz --verbose

will convert the arguments into the JSON format:

    {
        "align": true,
        "plot": true,
        "plot_mode": "xz",
        "verbose": true
    }

List arguments (--arg 1 2 3) are also supported.
To save the configuration, specify -o / --output.
'''


def main():
    import argcomplete
    basic_desc = "crappy configuration tool"
    lic = "(c) michael.grupp@tum.de"
    shared_parser = argparse.ArgumentParser(add_help=False)
    shared_parser.add_argument("--no_color", help="don't color output",
                               action="store_true")
    main_parser = argparse.ArgumentParser(
        description="%s %s" % (basic_desc, lic))
    sub_parsers = main_parser.add_subparsers(dest="subcommand")
    sub_parsers.required = True

    show_parser = sub_parsers.add_parser(
        "show", description="show configuration - %s" % lic,
        parents=[shared_parser])
    show_parser.add_argument(
        "config",
        help="optional config file to display (default: package settings)",
        nargs='?')
    show_parser.add_argument("--brief", help="show only the .json data",
                             action="store_true")

    set_parser = sub_parsers.add_parser(
        "set", description=SET_HELP, parents=[shared_parser],
        formatter_class=argparse.RawTextHelpFormatter)
    set_parser.add_argument("params", choices=list(
        DEFAULT_SETTINGS_DICT.keys()), nargs=argparse.REMAINDER,
                            help="parameters to set")
    set_parser.add_argument(
        "-c", "--config",
        help="optional config file (default: package settings)", default=None)
    set_parser.add_argument("-m", "--merge",
                            help="other config file to merge in (priority)",
                            default=None)
    set_parser.add_argument("--soft", help="do a soft-merge (no overwriting)",
                            action="store_true")

    gen_parser = sub_parsers.add_parser(
        "generate", description=GENERATE_HELP, parents=[shared_parser],
        formatter_class=argparse.RawTextHelpFormatter)
    gen_parser.add_argument("-o", "--out",
                            help="path for config file to generate")

    reset_parser = sub_parsers.add_parser(
        "reset", description="reset package settings - %s" % lic,
        parents=[shared_parser])
    reset_parser.add_argument("-y", help="acknowledge automatically",
                              action="store_true")

    argcomplete.autocomplete(main_parser)
    if len(sys.argv) > 1 and sys.argv[1] == "set":
        args, other_args = main_parser.parse_known_args()
        other_args = [arg for arg in sys.argv[2:] if not arg.startswith('-')]
    else:
        args, other_args = main_parser.parse_known_args()
    log.configure_logging()
    colorama.init()

    config = settings.DEFAULT_PATH
    if hasattr(args, "config"):
        if args.config:
            config = args.config

    if args.subcommand == "show":
        if not args.brief and not args.config:
            style = Style.BRIGHT if not args.no_color else Style.NORMAL
            doc_str = "\n".join(
                "{0}{1}{2}:\n{3}\n".format(style, k, Style.RESET_ALL, v[1])
                for k, v in sorted(DEFAULT_SETTINGS_DICT_DOC.items()))
            logger.info(doc_str)
            logger.info("{0}\n{1}\n{0}".format(SEP, config))
        show(config, colored=not args.no_color)
        if config == settings.DEFAULT_PATH and not args.brief:
            logger.info(SEP + "\nSee text above for parameter descriptions.")

    elif args.subcommand == "set":
        if not os.access(config, os.W_OK):
            logger.error("No permission to modify " + config)
            sys.exit()
        if other_args or args.merge:
            logger.info("{0}\nOld configuration:\n{0}".format(SEP))
            show(config, colored=not args.no_color)
            try:
                set_cfg(config, other_args)
            except ConfigError as e:
                logger.error(e)
                sys.exit(1)
            if args.merge:
                merge_json_union(config, args.merge, args.soft)
            logger.info(SEP + "\nNew configuration:\n" + SEP)
            show(config, colored=not args.no_color)
        else:
            logger.error("No configuration parameters given (see --help).")

    elif args.subcommand == "generate":
        if other_args:
            logger.info(
                "{0}\nParsed by argparse:\n{1}\n"
                "{0}\nWARNING:\nMake sure you use the 'long-style' -- options "
                "(e.g. --plot) if possible\nand no combined short '-' flags, "
                "(e.g. -vp)\n{0}".format(SEP, other_args))
            data = generate(other_args)
            log_info_dict_json(data, colored=not args.no_color)
            if args.out and user.check_and_confirm_overwrite(args.out):
                with open(args.out, 'w') as out:
                    out.write(json.dumps(data, indent=4, sort_keys=True))
            elif not args.out:
                logger.warning("\n(-o | --out) not specified - saving nothing")
        else:
            logger.error("No command line arguments given (see --help)")

    elif args.subcommand == "reset":
        if not os.access(config, os.W_OK):
            logger.error("No permission to modify" + config)
            sys.exit()
        if args.y or user.confirm(
                "Reset the package settings to the default settings? (y/n)"):
            settings.reset()
            logger.info("{0}\nPackage settings after reset:\n{0}".format(SEP))
            show(settings.DEFAULT_PATH, colored=not args.no_color)


if __name__ == '__main__':
    main()
